# ruff: noqa: SLF001
import datetime
import logging
import time
from collections import defaultdict
from collections.abc import Collection, Generator
from typing import TYPE_CHECKING, Any
from unittest import mock

import dagster._check as check
import pytest
from dagster._core.instance import DagsterInstance
from dagster._core.test_utils import environ
from dagster._serdes import serialize_value
from dagster._time import datetime_from_timestamp, get_current_datetime, get_current_timestamp
from dagster._utils.error import SerializableErrorInfo
from dagster_cloud import DagsterCloudAgentInstance
from dagster_cloud.api.dagster_cloud_api import (
    DagsterCloudUploadLocationData,
    DagsterCloudUploadWorkspaceEntry,
    UserCodeDeploymentType,
)
from dagster_cloud.execution.monitoring import CloudCodeServerHeartbeat, CloudCodeServerStatus
from dagster_cloud.pex.grpc.types import PexServerHandle
from dagster_cloud.workspace.user_code_launcher import (
    DagsterCloudGrpcServer,
    DagsterCloudUserCodeLauncher,
    ServerEndpoint,
    UserCodeLauncherEntry,
)
from dagster_cloud.workspace.user_code_launcher.user_code_launcher import (
    ACTUAL_ENTRIES_REFRESH_INTERVAL,
    CLEANUP_SERVER_GRACE_PERIOD_SECONDS,
    async_serialize_exceptions,
)
from dagster_cloud.workspace.user_code_launcher.utils import unique_resource_name
from dagster_cloud_cli.core.workspace import CodeLocationDeployData, PexMetadata
from dagster_shared.serdes.utils import create_snapshot_id
from freezegun import freeze_time

if TYPE_CHECKING:
    from dagster_cloud.workspace.user_code_launcher.user_code_launcher import (
        UserCodeLauncherEntryMap,
    )

ServerKey = tuple[str, str]
ServerHandle = str


class UserCodeTestLauncher(DagsterCloudUserCodeLauncher[ServerHandle]):
    def __init__(self, requires_images=True):
        self._servers: dict[ServerKey, set[ServerHandle]] = {}
        self._multipex_server_handles: dict[ServerKey, set[ServerHandle]] = {}

        self._pex_servers: dict[ServerHandle, list[PexServerHandle]] = defaultdict(list)

        self._agent_ids: dict[ServerHandle, str] = {}
        self._server_create_timestamps: dict[ServerHandle, float] = {}

        self._mocked_active_agent_ids: set[str] = set()

        self._server_create_counter = 0
        super().__init__()
        self.fail_next_add = False
        self.fail_next_add_after_server_added = set()

        self.fail_next_multipex_health_check = False

        self.fail_next_remove = False
        self.fail_graphql_writes = False
        self._requires_images = requires_images
        self._uploaded_entries = []

    def get_host_name(self, deployment_name, location_name, metadata):
        # pex_tag does not determine the hostname so we replicate that logic in this test
        metadata = metadata._replace(
            pex_metadata=PexMetadata(
                pex_tag="",
                python_version=metadata.pex_metadata.python_version
                if metadata.pex_metadata
                else None,
            )
        )
        return f"{deployment_name}_{location_name}_{create_snapshot_id(metadata)[0:6]}"

    def get_active_agent_ids(self) -> set[str]:
        return self._mocked_active_agent_ids

    @property
    def requires_images(self):
        return self._requires_images

    @property
    def servers(self) -> dict[ServerKey, set[ServerHandle]]:
        return self._servers.copy()

    @property
    def user_code_deployment_type(self) -> UserCodeDeploymentType:
        return UserCodeDeploymentType.UNKNOWN

    @property
    def multipex_server_handles(self) -> dict[ServerKey, set[ServerHandle]]:
        return self._multipex_server_handles.copy()

    @property
    def pex_servers(self) -> dict[str, list[PexServerHandle]]:
        return self._pex_servers.copy()

    def get_code_server_resource_limits(self, deployment_name, location_name):  # pyright: ignore[reportIncompatibleMethodOverride]
        return {}

    def _start_new_server_spinup(
        self,
        deployment_name: str,
        location_name: str,
        desired_entry: UserCodeLauncherEntry,
    ) -> DagsterCloudGrpcServer:
        metadata = desired_entry.code_location_deploy_data

        if self.fail_next_add:
            self.fail_next_add = False
            raise Exception("Server add failed")

        self._server_create_counter = self._server_create_counter + 1

        container_name = unique_resource_name(
            deployment_name, location_name, length_limit=50, sanitize_fn=None
        )

        endpoint = ServerEndpoint(
            host=self.get_host_name(deployment_name, location_name, metadata),
            port=4000 + self._server_create_counter,
            socket=None,
        )

        server_key = (deployment_name, location_name)

        if metadata.pex_metadata:
            if server_key not in self._servers:
                self._multipex_server_handles[server_key] = set()

            self._multipex_server_handles[server_key].add(container_name)

        else:
            if server_key not in self._servers:
                self._servers[server_key] = set()

            self._servers[server_key].add(container_name)
            self._agent_ids[container_name] = self._instance.instance_uuid
            self._server_create_timestamps[container_name] = get_current_timestamp()

        return DagsterCloudGrpcServer(container_name, endpoint, metadata)

    def get_agent_id_for_server(self, handle: ServerHandle) -> str | None:
        return self._agent_ids.get(handle)

    def get_server_create_timestamp(self, handle: ServerHandle) -> float | None:
        return self._server_create_timestamps.get(handle)

    def _list_server_handles(self) -> list[ServerHandle]:
        all_handles = set()
        for handles in self._servers.values():
            for handle in handles:
                all_handles.add(handle)
        return list(all_handles) + list(self._pex_servers.keys())

    @async_serialize_exceptions
    async def _wait_for_new_multipex_server(
        self,
        deployment_name: str,
        location_name: str,
        server_handle: ServerHandle,
        multipex_endpoint: ServerEndpoint,
    ):
        if location_name in self.fail_next_add_after_server_added:
            self.fail_next_add_after_server_added.remove(location_name)
            raise Exception("Server add failed after creating the server")

    async def _wait_for_new_server_ready(
        self,
        deployment_name: str,
        location_name: str,
        desired_entry: UserCodeLauncherEntry,
        server_handle: ServerHandle,
        server_endpoint: ServerEndpoint,
    ) -> None:
        if location_name in self.fail_next_add_after_server_added:
            self.fail_next_add_after_server_added.remove(location_name)
            raise Exception("Server add failed after creating the server")

    def _create_pex_server(
        self,
        deployment_name: str,
        location_name: str,
        desired_entry: UserCodeLauncherEntry,
        multipex_server: DagsterCloudGrpcServer,
    ):
        self._pex_servers[multipex_server.server_handle].append(
            PexServerHandle(
                deployment_name=deployment_name,
                location_name=location_name,
                metadata_update_timestamp=int(desired_entry.update_timestamp),
            )
        )
        self._agent_ids[multipex_server.server_handle] = self._instance.instance_uuid
        self._server_create_timestamps[multipex_server.server_handle] = get_current_timestamp()

    def _get_existing_pex_servers(
        self, deployment_name: str, location_name: str
    ) -> list[PexServerHandle]:
        multipex_server = self._multipex_servers.get((deployment_name, location_name))
        if not multipex_server:
            return []

        return self._pex_servers.get(multipex_server.server_handle, []).copy()

    def _remove_pex_server_handle(  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        self,
        _deployment_name,
        _location_name,
        server_handle: ServerHandle,
        _server_endpoint: ServerEndpoint,
        pex_server_handle: PexServerHandle,
    ) -> None:
        self._pex_servers[server_handle].remove(pex_server_handle)

    async def _get_upload_location_data(
        self,
        deployment_name: str,
        location_name: str,
        server: DagsterCloudGrpcServer,
    ) -> DagsterCloudUploadLocationData:
        return DagsterCloudUploadLocationData(
            upload_repository_datas=[],
            container_image="fake_image",
            executable_path="fake_executable",
        )

    def _check_running_multipex_server(self, _multipex_server: DagsterCloudGrpcServer):  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        if self.fail_next_multipex_health_check:
            self.fail_next_multipex_health_check = False
            raise Exception("can no longer reach multipex server")

    def _get_standalone_dagster_server_handles_for_location(
        self, deployment_name: str, location_name: str
    ) -> Collection[str]:
        return self._servers.get((deployment_name, location_name), set()).copy()

    def _get_multipex_server_handles_for_location(
        self, deployment_name: str, location_name: str
    ) -> Collection[str]:
        return self._multipex_server_handles.get((deployment_name, location_name), set()).copy()

    def _remove_server_handle(self, server_handle: ServerHandle) -> None:
        if self.fail_next_remove:
            self.fail_next_remove = False
            raise Exception("Server cleanup after update failed")

        for server_handles in self._servers.values():
            if server_handle in server_handles:
                server_handles.remove(server_handle)
        for server_handles in self._multipex_server_handles.values():
            if server_handle in server_handles:
                server_handles.remove(server_handle)

        self._pex_servers.pop(server_handle, None)
        self._agent_ids.pop(server_handle, None)
        self._server_create_timestamps.pop(server_handle, None)

        time.sleep(0.5)

    def _update_workspace_entry(self, _deployment_name: str, workspace_entry, server_or_error):  # pyright: ignore[reportIncompatibleMethodOverride], fix me!
        if self.fail_graphql_writes:
            raise Exception("Update workspace entry failed")

        self._uploaded_entries.append(workspace_entry)

    def _resolve_image(self, metadata: CodeLocationDeployData) -> str | None:
        # use the pex python version if the image is missing, to simulate the behavior in serverless
        if metadata.image:
            return metadata.image
        if metadata.pex_metadata and metadata.pex_metadata.python_version:
            return "base-image-for:" + metadata.pex_metadata.python_version

        return None

    @property
    def uploaded_entries(self) -> list[DagsterCloudUploadWorkspaceEntry]:
        return self._uploaded_entries.copy()

    @property
    def uploaded_locations(self) -> list[str]:
        return [entry.location_name for entry in self._uploaded_entries]

    def wait_for_reconcile(self):
        with self._metadata_lock:
            start_count = self._reconcile_count

        # Wait for reconcile
        start_time = time.time()
        timeout = 5

        while True:
            with self._metadata_lock:
                new_count = self._reconcile_count

            if new_count > start_count:
                break

            if time.time() - start_time > timeout:
                raise Exception("Timed out waiting for reconcile")
            time.sleep(1)

    def assert_reconciliation(
        self,
        expected_entry_map: dict[tuple[str, str], UserCodeLauncherEntry],
        expected_errors: dict[tuple[Any, ...], str] | None = None,
        run_reconcile_thread: bool = True,
    ):
        if run_reconcile_thread:
            self.wait_for_reconcile()

        expected_errors = check.opt_dict_param(
            expected_errors, "expected_errors", key_type=tuple, value_type=str
        )
        expected_metadata = {
            location: entry.code_location_deploy_data
            for location, entry in expected_entry_map.items()
        }

        grpc_endpoints = self.get_grpc_endpoints()
        grpc_map = {
            key: grpc_endpoint.host
            for key, grpc_endpoint in grpc_endpoints.items()
            if not isinstance(grpc_endpoint, SerializableErrorInfo)
        }

        error_map = {
            key: grpc_endpoint
            for key, grpc_endpoint in grpc_endpoints.items()
            if isinstance(grpc_endpoint, SerializableErrorInfo)
        }

        metadata_map = {}
        for key, metadata in expected_metadata.items():
            deployment_name, location_name = key
            metadata_map[key] = self.get_host_name(deployment_name, location_name, metadata)

        assert grpc_map == metadata_map, str(error_map)

        assert expected_errors.keys() == error_map.keys()
        for error in error_map:
            assert expected_errors[error] in str(error_map[error])

    def run_launcher(self):
        raise NotImplementedError()


def _check_update(
    user_code_launcher,
    entry_map: dict[tuple[str, str], UserCodeLauncherEntry],
    expected_entry_map: dict[tuple[str, str], UserCodeLauncherEntry] | None = None,
    agent_error_map=None,
    control_plane_error_locations: set[tuple[str, str]] | None = None,
):
    expected_entry_map = expected_entry_map if expected_entry_map is not None else entry_map
    user_code_launcher.update_grpc_metadata(
        entry_map,
        control_plane_error_locations=control_plane_error_locations or set(),
        control_plane_outdated_locations=set(),
    )
    user_code_launcher.assert_reconciliation(expected_entry_map, agent_error_map)


@pytest.fixture
def user_code_launcher(
    caplog: pytest.LogCaptureFixture, agent_instance: DagsterCloudAgentInstance
) -> Generator[UserCodeTestLauncher, None, None]:
    caplog.set_level(logging.INFO)
    with UserCodeTestLauncher() as user_code_launcher:
        user_code_launcher.register_instance(agent_instance)
        yield user_code_launcher


@pytest.fixture
def user_code_launcher_no_images(
    agent_instance: DagsterCloudAgentInstance,
) -> Generator[UserCodeTestLauncher, None, None]:
    with UserCodeTestLauncher(requires_images=False) as user_code_launcher:
        user_code_launcher.register_instance(agent_instance)
        yield user_code_launcher


def test_server_cleanup_failure(
    agent_instance: DagsterCloudAgentInstance, caplog: pytest.LogCaptureFixture
):
    with UserCodeTestLauncher() as user_code_launcher:
        user_code_launcher.register_instance(agent_instance)
        user_code_launcher.start(run_reconcile_thread=False)
        now = time.time()
        server_one = UserCodeLauncherEntry(
            CodeLocationDeployData("test_image:tag1", python_file="foo.py"),
            now,
        )

        user_code_launcher.update_grpc_metadata(
            {("dep1", "location1"): server_one},
            control_plane_error_locations=set(),
            control_plane_outdated_locations=set(),
        )
        user_code_launcher.get_grpc_server_heartbeats()
        user_code_launcher.reconcile()
        user_code_launcher.get_grpc_server_heartbeats()

        user_code_launcher.fail_next_remove = True
        # Ensure that cleaning up servers doesnt cause program exit
        user_code_launcher._cleanup_servers(
            {agent_instance.instance_uuid}, include_own_servers=True
        )

        # Ensure that an error actually occurred, and we logged it.
        assert caplog.text.count("Error cleaning up server") == 1


def test_server_cleanup_parallelized(agent_instance: DagsterCloudAgentInstance):
    with UserCodeTestLauncher() as user_code_launcher:
        user_code_launcher.register_instance(agent_instance)
        user_code_launcher.start(run_reconcile_thread=False)
        now = time.time()
        server = UserCodeLauncherEntry(
            CodeLocationDeployData("test_image:tag1", python_file="foo.py"),
            now,
        )
        user_code_launcher.update_grpc_metadata(
            {(f"dep{i}", f"location{i}"): server for i in range(20)},
            control_plane_error_locations=set(),
            control_plane_outdated_locations=set(),
        )

        user_code_launcher.get_grpc_server_heartbeats()
        user_code_launcher.reconcile()
        user_code_launcher.get_grpc_server_heartbeats()

        now = time.time()
        user_code_launcher._cleanup_servers(
            {agent_instance.instance_uuid}, include_own_servers=True
        )
        assert time.time() - now < 5


def test_multiple_agents_grpc_spindown(agent_instance: DagsterCloudAgentInstance):
    with UserCodeTestLauncher() as user_code_launcher:
        active_agent_ids = {"different-agent-id", agent_instance.instance_uuid}
        user_code_launcher._mocked_active_agent_ids = active_agent_ids
        user_code_launcher.register_instance(agent_instance)
        user_code_launcher.start(run_reconcile_thread=False)
        now = time.time()

        one_hour_ago = now - CLEANUP_SERVER_GRACE_PERIOD_SECONDS - 1

        server_one = UserCodeLauncherEntry(
            CodeLocationDeployData("test_image:tag1", python_file="foo.py"),
            now,
        )
        server_two = UserCodeLauncherEntry(
            CodeLocationDeployData("test_image:tag2", python_file="foo.py"), now
        )
        server_three = UserCodeLauncherEntry(
            CodeLocationDeployData("test_image:tag3", python_file="foo.py"), now
        )
        server_four = UserCodeLauncherEntry(
            CodeLocationDeployData("test_image:tag4", python_file="foo.py"), now
        )

        pex_server_one = UserCodeLauncherEntry(
            CodeLocationDeployData(
                "test_image:tag1",
                python_file="foo.py",
                pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
            ),
            now,
        )
        pex_server_two = UserCodeLauncherEntry(
            CodeLocationDeployData(
                "test_image:tag2",
                python_file="foo.py",
                pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
            ),
            now,
        )
        pex_server_three = UserCodeLauncherEntry(
            CodeLocationDeployData(
                "test_image:tag3",
                python_file="foo.py",
                pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
            ),
            now,
        )
        pex_server_four = UserCodeLauncherEntry(
            CodeLocationDeployData(
                "test_image:tag4",
                python_file="foo.py",
                pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
            ),
            now,
        )

        with freeze_time(datetime.datetime.fromtimestamp(one_hour_ago, tz=datetime.timezone.utc)):
            entry_map: UserCodeLauncherEntryMap = {
                ("dep1", "location1"): server_one,
                ("dep2", "location2"): server_two,
                ("dep3", "location3"): server_three,
                ("dep1", "location5"): pex_server_one,
                ("dep2", "location6"): pex_server_two,
                ("dep3", "location7"): pex_server_three,
            }

            user_code_launcher.update_grpc_metadata(
                entry_map,
                control_plane_error_locations=set(),
                control_plane_outdated_locations=set(),
            )
            user_code_launcher.get_grpc_server_heartbeats()
            user_code_launcher.reconcile()
            user_code_launcher.get_grpc_server_heartbeats()

            assert len(user_code_launcher._agent_ids.keys()) == 6

        with freeze_time(datetime.datetime.fromtimestamp(now, tz=datetime.timezone.utc)):
            entry_map[("dep4", "location4")] = server_four
            entry_map[("dep4", "location8")] = pex_server_four

            user_code_launcher.update_grpc_metadata(
                entry_map,
                control_plane_error_locations=set(),
                control_plane_outdated_locations=set(),
            )
            user_code_launcher.get_grpc_server_heartbeats()
            user_code_launcher.reconcile()
            user_code_launcher.get_grpc_server_heartbeats()

            assert len(user_code_launcher._agent_ids.keys()) == 8
            # Since all servers were created by this agent instance, both should be able to clean up.
            for key in user_code_launcher._agent_ids.keys():
                assert user_code_launcher._can_cleanup_server(
                    key, active_agent_ids, include_own_servers=True
                )

                # even if the active_agent_ids are None due to the graphql query failing
                assert user_code_launcher._can_cleanup_server(
                    key, active_agent_ids=None, include_own_servers=True
                )

            handle_active_agent = [
                key
                for key in user_code_launcher._agent_ids.keys()
                if key.startswith("location1-dep1")
            ].pop()
            pex_handle_active_agent = [
                key
                for key in user_code_launcher._agent_ids.keys()
                if key.startswith("location5-dep1")
            ].pop()

            handle_inactive_agent_not_recent = [
                key
                for key in user_code_launcher._agent_ids.keys()
                if key.startswith("location2-dep2")
            ].pop()
            pex_handle_inactive_agent_not_recent = [
                key
                for key in user_code_launcher._agent_ids.keys()
                if key.startswith("location6-dep2")
            ].pop()

            handle_inactive_agent_recent = [
                key
                for key in user_code_launcher._agent_ids.keys()
                if key.startswith("location4-dep4")
            ].pop()
            pex_handle_inactive_agent_recent = [
                key
                for key in user_code_launcher._agent_ids.keys()
                if key.startswith("location8-dep4")
            ].pop()

            # Induce a state where one of the grpc servers was created by a different agent which is active.
            user_code_launcher._agent_ids[handle_active_agent] = "different-agent-id"
            user_code_launcher._agent_ids[pex_handle_active_agent] = "different-agent-id"

            assert not user_code_launcher._can_cleanup_server(
                handle_active_agent, active_agent_ids, include_own_servers=True
            )
            assert not user_code_launcher._can_cleanup_server(
                pex_handle_active_agent,
                active_agent_ids,
                include_own_servers=True,
            )

            # Induce a state where one of the grpc servers was created by a different agent which is inactive.
            user_code_launcher._agent_ids[handle_inactive_agent_recent] = "inactive-agent-id"
            user_code_launcher._agent_ids[pex_handle_inactive_agent_recent] = "inactive-agent-id"

            # recent ones still cannot be cleaned up
            assert not user_code_launcher._can_cleanup_server(
                handle_inactive_agent_recent, active_agent_ids, include_own_servers=True
            )
            assert not user_code_launcher._can_cleanup_server(
                pex_handle_inactive_agent_recent, active_agent_ids, include_own_servers=True
            )

            # non-recent ones can
            user_code_launcher._agent_ids[handle_inactive_agent_not_recent] = "inactive-agent-id"
            user_code_launcher._agent_ids[pex_handle_inactive_agent_not_recent] = (
                "inactive-agent-id"
            )

            assert user_code_launcher._can_cleanup_server(
                handle_inactive_agent_not_recent, active_agent_ids, include_own_servers=True
            )
            assert user_code_launcher._can_cleanup_server(
                pex_handle_inactive_agent_not_recent, active_agent_ids, include_own_servers=True
            )

            # If the active agent IDs are None (due to the graphql query failing) they can no longer be cleaned up
            assert not user_code_launcher._can_cleanup_server(
                handle_inactive_agent_not_recent, active_agent_ids=None, include_own_servers=True
            )
            assert not user_code_launcher._can_cleanup_server(
                pex_handle_inactive_agent_not_recent,
                active_agent_ids=None,
                include_own_servers=True,
            )

            # Perform cleanup of grpc servers
            user_code_launcher._graceful_cleanup_servers(include_own_servers=True)
            remaining_handles = set()
            for handles in user_code_launcher._servers.values():
                for handle in handles:
                    remaining_handles.add(handle)
            for handle in user_code_launcher._pex_servers.keys():
                remaining_handles.add(handle)
            # Ensure that the handle created by this agent, and the handle created by
            # an inactive agent were both cleaned up.
            assert remaining_handles == {
                handle_active_agent,
                pex_handle_active_agent,
                handle_inactive_agent_recent,
                pex_handle_inactive_agent_recent,
            }


def test_reconcile_loop_does_cleanup(agent_instance: DagsterCloudAgentInstance):
    with UserCodeTestLauncher() as user_code_launcher:
        active_agent_ids = {"different-agent-id", agent_instance.instance_uuid}
        user_code_launcher._mocked_active_agent_ids = active_agent_ids
        user_code_launcher.register_instance(agent_instance)
        user_code_launcher.start(run_reconcile_thread=False)
        now = time.time()
        with freeze_time(datetime_from_timestamp(now)):
            server_one = UserCodeLauncherEntry(
                CodeLocationDeployData("test_image:tag1", python_file="foo.py"),
                now,
            )
            server_two = UserCodeLauncherEntry(
                CodeLocationDeployData("test_image:tag2", python_file="foo.py"), now
            )
            server_three = UserCodeLauncherEntry(
                CodeLocationDeployData("test_image:tag3", python_file="foo.py"), now
            )
            entry_map: UserCodeLauncherEntryMap = {
                ("dep1", "location1"): server_one,
                ("dep2", "location2"): server_two,
                ("dep3", "location3"): server_three,
            }
            user_code_launcher.update_grpc_metadata(
                entry_map,
                control_plane_error_locations=set(),
                control_plane_outdated_locations=set(),
            )
            user_code_launcher.reconcile()

            assert len(user_code_launcher._list_server_handles()) == 3

            handles = list(user_code_launcher._agent_ids.keys())
            assert len(handles) == 3
            # Induce a state where two of the grpc servers was created by a different agent which is inactive.
            for i in range(2):
                user_code_launcher._agent_ids[handles[i]] = "inactive-agent-id"

            assert len(user_code_launcher._servers) == 3
            user_code_launcher.reconcile()

            # Still 3 until the right amount of time passes
            assert len(user_code_launcher._list_server_handles()) == 3

        with freeze_time(
            datetime_from_timestamp(
                now
                + max(
                    user_code_launcher._cleanup_server_check_interval(),
                    CLEANUP_SERVER_GRACE_PERIOD_SECONDS,
                )
                + 1
            )
        ):
            user_code_launcher.reconcile()
            # The two servers have now been cleaned up
            assert len(user_code_launcher._list_server_handles()) == 1


def test_recovery_after_failed_redeploy(user_code_launcher: UserCodeTestLauncher):
    """Test the self-healing recovery mechanism for failed non-upload redeploys.

    Scenario:
    1. Agent deploys a code server successfully.
    2. _refresh_actual_entries detects the server moved to an error state (e.g.
       list_repositories returns an error) and triggers a recovery redeploy.
    3. The redeploy fails transiently - agent is now stuck because actual == desired
       but _grpc_servers has an error.
    4. On the next workspace query, the server reports no error (hasLoadError=False).
       The recovery mechanism detects the mismatch and triggers another redeploy.
    5. The redeploy succeeds and the location is healthy again.
    """
    user_code_launcher.start(run_reconcile_thread=False)

    now = get_current_datetime()
    now_timestamp = now.timestamp()

    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now_timestamp
    )
    metadata_map: UserCodeLauncherEntryMap = {("dep1", "location1"): server_one}

    # Step 1: Deploy successfully
    with freeze_time(now):
        user_code_launcher.update_grpc_metadata(
            metadata_map,
            control_plane_error_locations=set(),
            control_plane_outdated_locations=set(),
        )
        user_code_launcher.reconcile()
        user_code_launcher.assert_reconciliation(
            expected_entry_map=metadata_map, run_reconcile_thread=False
        )
        initial_servers = user_code_launcher.get_grpc_endpoints()
        assert isinstance(initial_servers[("dep1", "location1")], ServerEndpoint)

    # Step 2: _refresh_actual_entries detects server moved to error state.
    # The recovery redeploy FAILS because of a transient error.
    frozen_datetime = now + datetime.timedelta(seconds=ACTUAL_ENTRIES_REFRESH_INTERVAL + 1)
    with freeze_time(frozen_datetime):
        # Make list_repositories return an error so _refresh_actual_entries triggers recovery
        mock_method = mock.patch.object(
            initial_servers[("dep1", "location1")],
            "create_client",
            return_value=mock.MagicMock(
                list_repositories=lambda: serialize_value(
                    SerializableErrorInfo(
                        message="test error", stack=[], cls_name="TestError", cause=None
                    )
                )
            ),
        )
        mock_method.start()

        # Make the redeploy fail
        user_code_launcher.fail_next_add = True

        # The control plane doesn't know about this error — the agent's local
        # health check (list_repositories) is what detects it.
        user_code_launcher.update_grpc_metadata(
            metadata_map,
            control_plane_error_locations=set(),
            control_plane_outdated_locations=set(),
        )
        user_code_launcher.reconcile()

        mock_method.stop()

        # Step 3: Agent is stuck - actual == desired but _grpc_servers has an error
        servers = user_code_launcher.get_grpc_endpoints()
        assert isinstance(servers[("dep1", "location1")], SerializableErrorInfo)
        assert ("dep1", "location1") in user_code_launcher._actual_entries

    # Step 4: Next workspace query - server reports no error (hasLoadError=False).
    # Recovery mechanism runs inside _refresh_actual_entries every 30s.
    frozen_datetime2 = frozen_datetime + datetime.timedelta(
        seconds=ACTUAL_ENTRIES_REFRESH_INTERVAL + 1
    )
    with freeze_time(frozen_datetime2):
        user_code_launcher.update_grpc_metadata(
            metadata_map,
            control_plane_error_locations=set(),
            control_plane_outdated_locations=set(),
        )
        user_code_launcher.reconcile()

        # Step 5: Location should be healthy again
        recovered_servers = user_code_launcher.get_grpc_endpoints()
        assert isinstance(recovered_servers[("dep1", "location1")], ServerEndpoint)


def test_redeploy_when_server_moves_to_error_state(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start(run_reconcile_thread=False)

    now = get_current_datetime()
    now_timestamp = now.timestamp()

    user_code_launcher.fail_next_add_after_server_added = {"location1"}

    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now_timestamp
    )
    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="bar.py"), now_timestamp
    )
    server_three = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag3", python_file="baz.py"), now_timestamp
    )
    metadata_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_one,
        ("dep1", "location2"): server_two,
        ("dep1", "location3"): server_three,
    }
    with freeze_time(now):
        # all three locations are 'outdated' (in the process of being updated),
        # but this agent is not being asked to upload the results for any of them
        user_code_launcher.update_grpc_metadata(
            metadata_map,
            control_plane_error_locations=set(),
            control_plane_outdated_locations={
                ("dep1", "location1"),
                ("dep1", "location2"),
                ("dep1", "location3"),
            },
        )
        user_code_launcher.reconcile()
        user_code_launcher.assert_reconciliation(
            expected_entry_map={
                ("dep1", "location2"): server_two,
                ("dep1", "location3"): server_three,
            },
            expected_errors={("dep1", "location1"): "Server add failed after creating the server"},
            run_reconcile_thread=False,
        )

        initial_reconciled_servers = user_code_launcher.get_grpc_endpoints()

        assert isinstance(initial_reconciled_servers[("dep1", "location1")], SerializableErrorInfo)
        assert isinstance(initial_reconciled_servers[("dep1", "location2")], ServerEndpoint)
        assert isinstance(initial_reconciled_servers[("dep1", "location3")], ServerEndpoint)

        initial_actual_entries = user_code_launcher._actual_entries

    frozen_datetime = now + datetime.timedelta(seconds=ACTUAL_ENTRIES_REFRESH_INTERVAL + 1)
    with freeze_time(frozen_datetime):
        # Make one server return an error from its list_repositories() call
        mock_method = mock.patch.object(
            initial_reconciled_servers[("dep1", "location2")],
            "create_client",
            return_value=mock.MagicMock(
                list_repositories=lambda: serialize_value(
                    SerializableErrorInfo(
                        message="test error", stack=[], cls_name="TestError", cause=None
                    )
                )
            ),
        )
        mock_method.start()

        user_code_launcher.reconcile()

        # entries are the same because the new code location is redeployed, but the new server is different
        assert user_code_launcher._actual_entries == initial_actual_entries

        new_servers = user_code_launcher.get_grpc_endpoints()
        # locations already in an error unchanged
        assert (
            new_servers[("dep1", "location1")] == initial_reconciled_servers[("dep1", "location1")]
        )
        assert (
            new_servers[("dep1", "location2")] != initial_reconciled_servers[("dep1", "location2")]
        )
        assert (
            new_servers[("dep1", "location3")] == initial_reconciled_servers[("dep1", "location3")]
        )


def test_refresh_actual_entries_multipex(
    user_code_launcher: UserCodeTestLauncher,
):
    """Test that _refresh_actual_entries handles dictionary mutation during iteration."""
    user_code_launcher.start(run_reconcile_thread=False)

    now = time.time()

    # Set up multiple pex locations
    pex_server_one = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now,
    )
    pex_server_two = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="bar.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now,
    )

    metadata_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): pex_server_one,
        ("dep1", "location2"): pex_server_two,
    }

    user_code_launcher.update_grpc_metadata(
        metadata_map, control_plane_error_locations=set(), control_plane_outdated_locations=set()
    )
    user_code_launcher.reconcile()

    # Verify both locations are set up with multipex servers
    assert ("dep1", "location1") in user_code_launcher._multipex_servers
    assert ("dep1", "location2") in user_code_launcher._multipex_servers
    assert ("dep1", "location1") in user_code_launcher._actual_entries
    assert ("dep1", "location2") in user_code_launcher._actual_entries

    # Clear pex servers for location1 to simulate them disappearing.
    # This will cause _refresh_actual_entries to call _trigger_recovery_server_restart
    # which removes the entry from _multipex_servers during iteration.
    multipex_server_1 = user_code_launcher._multipex_servers[("dep1", "location1")]
    multipex_server_2 = user_code_launcher._multipex_servers[("dep1", "location2")]
    user_code_launcher._pex_servers[multipex_server_1.server_handle] = []

    # Mock create_multipex_client to return a mock that succeeds on ping,
    # so _refresh_actual_entries proceeds past the health check to the pex server check
    mock_multipex_client = mock.MagicMock()
    with (
        mock.patch.object(
            multipex_server_1.server_endpoint,
            "create_multipex_client",
            return_value=mock_multipex_client,
        ),
        mock.patch.object(
            multipex_server_2.server_endpoint,
            "create_multipex_client",
            return_value=mock_multipex_client,
        ),
    ):
        user_code_launcher._refresh_actual_entries(metadata_map, set(), set(), set())

    # location1 should have been removed from _multipex_servers by _trigger_recovery_server_restart
    assert ("dep1", "location1") not in user_code_launcher._multipex_servers
    # location2 should still be present
    assert ("dep1", "location2") in user_code_launcher._multipex_servers


def test_redeploy_unavailable_server(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start(run_reconcile_thread=False)

    now = get_current_datetime()
    now_timestamp = now.timestamp()

    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now_timestamp
    )
    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="bar.py"), now_timestamp
    )

    metadata_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_one,
        ("dep1", "location2"): server_two,
    }

    with freeze_time(now):
        user_code_launcher.update_grpc_metadata(
            metadata_map,
            control_plane_error_locations=set(),
            control_plane_outdated_locations=set(),
        )
        user_code_launcher.reconcile()
        user_code_launcher.assert_reconciliation(metadata_map, None, run_reconcile_thread=False)

        reconciled_servers = user_code_launcher.get_grpc_endpoints()

        assert ("dep1", "location1") in reconciled_servers and (
            "dep1",
            "location2",
        ) in reconciled_servers

    # Both servers fail their ping() check (since they are not actually running)
    frozen_datetime = now + datetime.timedelta(seconds=ACTUAL_ENTRIES_REFRESH_INTERVAL + 1)
    with freeze_time(frozen_datetime):
        user_code_launcher.reconcile()

        assert user_code_launcher._first_unavailable_times == {
            ("dep1", "location1"): frozen_datetime.timestamp(),
            ("dep1", "location2"): frozen_datetime.timestamp(),
        }

        first_failure_time = frozen_datetime

        assert reconciled_servers == user_code_launcher.get_grpc_endpoints()

    # Make one server recover on the next check
    frozen_datetime = frozen_datetime + datetime.timedelta(
        seconds=ACTUAL_ENTRIES_REFRESH_INTERVAL + 1
    )
    with freeze_time(frozen_datetime):
        # Make one server recover its ping() check (by returning a Mock that passes ping() without raising an exception)
        mock_method = mock.patch.object(
            reconciled_servers[("dep1", "location1")],
            "create_client",
            return_value=mock.MagicMock(),
        )
        mock_method.start()
        user_code_launcher.reconcile()

        # Other location is still tracking
        assert user_code_launcher._first_unavailable_times == {
            ("dep1", "location2"): first_failure_time.timestamp(),
        }

    frozen_datetime = frozen_datetime + datetime.timedelta(
        user_code_launcher._server_process_startup_timeout + 1
    )
    with freeze_time(frozen_datetime):
        user_code_launcher.reconcile()

        # New server has been created since the server stayed unreachable for long enough
        new_servers = user_code_launcher.get_grpc_endpoints()

        assert new_servers[("dep1", "location1")] == reconciled_servers[("dep1", "location1")]
        assert new_servers[("dep1", "location2")] != reconciled_servers[("dep1", "location2")]


def test_heartbeats(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start(run_reconcile_thread=False)
    assert user_code_launcher.get_grpc_server_heartbeats() == {}

    now = time.time()

    # Add a server
    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )
    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="foo.py"), now
    )

    server_three = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag3", python_file="foo.py"), now + 1
    )

    entry_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_one,
        ("dep2", "location2"): server_two,
    }

    user_code_launcher.update_grpc_metadata(
        entry_map, control_plane_error_locations=set(), control_plane_outdated_locations=set()
    )

    assert user_code_launcher.get_grpc_server_heartbeats() == {
        "dep1": [
            CloudCodeServerHeartbeat(
                "location1",
                server_status=CloudCodeServerStatus.STARTING,
                error=None,
                metadata={},
            )
        ],
        "dep2": [
            CloudCodeServerHeartbeat(
                "location2",
                server_status=CloudCodeServerStatus.STARTING,
                error=None,
                metadata={},
            )
        ],
    }

    user_code_launcher.reconcile()

    assert user_code_launcher.get_grpc_server_heartbeats() == {
        "dep1": [
            CloudCodeServerHeartbeat(
                "location1",
                server_status=CloudCodeServerStatus.RUNNING,
                error=None,
                metadata={},
            )
        ],
        "dep2": [
            CloudCodeServerHeartbeat(
                "location2",
                server_status=CloudCodeServerStatus.RUNNING,
                error=None,
                metadata={},
            )
        ],
    }
    # Test recovery after adding a server
    user_code_launcher.fail_next_add = True

    entry_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_three,
        ("dep2", "location2"): server_two,
    }

    user_code_launcher.update_grpc_metadata(
        entry_map, control_plane_error_locations=set(), control_plane_outdated_locations=set()
    )
    assert user_code_launcher.get_grpc_server_heartbeats() == {
        "dep1": [
            CloudCodeServerHeartbeat(
                "location1",
                server_status=CloudCodeServerStatus.RUNNING,
                error=None,
                metadata={},
            )
        ],
        "dep2": [
            CloudCodeServerHeartbeat(
                "location2",
                server_status=CloudCodeServerStatus.RUNNING,
                error=None,
                metadata={},
            )
        ],
    }

    user_code_launcher.reconcile()

    heartbeats_with_error = user_code_launcher.get_grpc_server_heartbeats()

    assert len(heartbeats_with_error) == 2
    assert len(heartbeats_with_error["dep1"]) == 1
    assert heartbeats_with_error["dep1"][0].location_name == "location1"
    assert heartbeats_with_error["dep1"][0].server_status == CloudCodeServerStatus.FAILED
    assert "Server add failed" in str(heartbeats_with_error["dep1"][0].error)

    assert heartbeats_with_error["dep2"] == [
        CloudCodeServerHeartbeat(
            "location2",
            server_status=CloudCodeServerStatus.RUNNING,
            error=None,
            metadata={},
        )
    ]

    # Test with truncation
    with environ({"DAGSTER_CLOUD_CODE_SERVER_HEARTBEAT_ERROR_SIZE_LIMIT": "15"}):
        heartbeats_with_error = user_code_launcher.get_grpc_server_heartbeats()
        assert len(heartbeats_with_error) == 2
        assert len(heartbeats_with_error["dep1"]) == 1
        assert heartbeats_with_error["dep1"][0].location_name == "location1"
        assert heartbeats_with_error["dep1"][0].server_status == CloudCodeServerStatus.FAILED
        assert "Exception: Serv (TRUNCATED)" in str(heartbeats_with_error["dep1"][0].error)


def test_updating_metadata_without_timestamp_does_not_update_code_server(
    user_code_launcher: UserCodeTestLauncher,
):
    user_code_launcher.start()

    now = time.time()

    assert user_code_launcher.get_active_grpc_server_handles() == []

    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )
    _check_update(user_code_launcher, {("dep1", "location1"): server_one})

    assert len(user_code_launcher.get_active_grpc_server_handles()) == 1

    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1", python_file="foo.py", cloud_context_env={"FOO": "BAR"}
        ),
        now,
    )

    # since only cloud context env changed, nothing updates
    _check_update(
        user_code_launcher,
        {("dep1", "location1"): server_two},
        expected_entry_map={("dep1", "location1"): server_one},
    )


def test_reconciliation(user_code_launcher: UserCodeTestLauncher, caplog: pytest.LogCaptureFixture):
    user_code_launcher.start()

    now = time.time()

    caplog.clear()

    assert user_code_launcher.get_active_grpc_server_handles() == []

    # Add a server
    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )
    _check_update(user_code_launcher, {("dep1", "location1"): server_one})

    assert (
        caplog.text.count(
            f"Reconciling to reach {{(dep1, location1, {now})}}. To add: {{(dep1, location1,"
            f" {now})}}. To update: {{}}. To remove: {{}}. To upload: {{}}."
        )
        == 1
    )

    assert len(user_code_launcher.get_active_grpc_server_handles()) == 1

    caplog.clear()

    # Update a server to a new image
    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="foo.py"), now + 1
    )
    _check_update(user_code_launcher, {("dep1", "location1"): server_two})

    assert (
        caplog.text.count(
            f"Reconciling to reach {{(dep1, location1, {server_two.update_timestamp})}}. To add: {{}}. To update: {{(dep1,"
            f" location1, {server_two.update_timestamp})}}. To remove: {{}}. To upload: {{}}."
        )
        == 1
    )

    assert len(user_code_launcher.get_active_grpc_server_handles()) == 1

    caplog.clear()

    # Add a second server
    server_three = UserCodeLauncherEntry(
        CodeLocationDeployData("test_another_image:tag3", python_file="bar.py"), now + 2
    )
    _check_update(
        user_code_launcher,
        {
            ("dep1", "location1"): server_two,
            ("dep1", "location2"): server_three,
        },
    )

    assert (
        caplog.text.count(
            f"Reconciling to reach {{(dep1, location1, {server_two.update_timestamp}), (dep1, location2, {server_three.update_timestamp})}}. To add:"
            f" {{(dep1, location2, {server_three.update_timestamp})}}. To update: {{}}. To remove: {{}}. To upload: {{}}."
        )
        == 1
    )

    assert len(user_code_launcher.get_active_grpc_server_handles()) == 2

    caplog.clear()

    # Remove a server
    _check_update(user_code_launcher, {("dep1", "location2"): server_three})

    assert (
        caplog.text.count(
            f"Reconciling to reach {{(dep1, location2, {server_three.update_timestamp})}}. To add: {{}}. To update: {{}}. To"
            f" remove: {{(dep1, location1, {server_two.update_timestamp})}}. To upload: {{}}."
        )
        == 1
    )

    assert len(user_code_launcher.get_active_grpc_server_handles()) == 1

    # Since we didn't specify any locations should be uploaded, none were
    assert user_code_launcher.uploaded_locations == []


def test_new_timestamp_refreshes_server(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start()

    now = time.time()

    metadata = CodeLocationDeployData("test_image:tag1", python_file="foo.py")

    server_one = UserCodeLauncherEntry(metadata, now)
    _check_update(user_code_launcher, {("dep1", "location1"): server_one})

    current_endpoint = user_code_launcher.get_grpc_endpoint("dep1", "location1")
    _check_update(user_code_launcher, {("dep1", "location1"): server_one})
    assert user_code_launcher.get_grpc_endpoint("dep1", "location1").port == current_endpoint.port

    # Increasing timestamp causes the server to be reloaded
    _check_update(
        user_code_launcher,
        {("dep1", "location1"): UserCodeLauncherEntry(metadata, now + 1)},
    )
    assert (
        user_code_launcher.get_grpc_endpoint(
            "dep1",
            "location1",
        ).port
        != current_endpoint.port
    )

    assert user_code_launcher.uploaded_locations == []


def test_graphql_write_failure(
    caplog: pytest.LogCaptureFixture, user_code_launcher: UserCodeTestLauncher
):
    user_code_launcher.fail_graphql_writes = True

    user_code_launcher.start()

    now = time.time()

    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )
    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="bar.py"), now
    )

    metadata_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_one,
        ("dep1", "location2"): server_two,
    }

    user_code_launcher.add_upload_metadata_for_deployment(
        "dep1",
        upload_metadata=metadata_map,
        control_plane_error_locations=set(),
        control_plane_outdated_locations=set(metadata_map.keys()),
    )
    user_code_launcher.wait_for_reconcile()

    # graphql fails, but the metadata updates
    user_code_launcher.assert_reconciliation(metadata_map)

    assert caplog.text.count("Fetching metadata for ") == 2
    assert caplog.text.count("Error while writing location data") == 2

    caplog.clear()

    user_code_launcher.wait_for_reconcile()
    assert caplog.text.count("Error while writing location data") == 0

    assert user_code_launcher.uploaded_locations == []


def test_add_upload_metadata_retains_other_deployments(
    user_code_launcher: UserCodeTestLauncher,
):
    """Test that add_upload_metadata_for_deployment only replaces error/outdated
    state for the given deployment, preserving state for other deployments.
    """
    user_code_launcher.start(run_reconcile_thread=False)

    now = time.time()
    dep1_loc = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )
    dep2_loc = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="bar.py"), now
    )

    # Set up initial state with two deployments via update_grpc_metadata
    # (which replaces all state)
    user_code_launcher.update_grpc_metadata(
        {("dep1", "location1"): dep1_loc, ("dep2", "location1"): dep2_loc},
        control_plane_error_locations={("dep2", "location1")},
        control_plane_outdated_locations={("dep1", "location1")},
    )

    assert user_code_launcher._control_plane_error_locations == {("dep2", "location1")}
    assert user_code_launcher._control_plane_outdated_locations == {("dep1", "location1")}

    # Now add upload metadata for dep1 only — dep2's error/outdated state must be retained
    dep1_loc_updated = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1_v2", python_file="foo.py"), now + 1
    )
    user_code_launcher.add_upload_metadata_for_deployment(
        "dep1",
        upload_metadata={("dep1", "location1"): dep1_loc_updated},
        control_plane_error_locations={("dep1", "location1")},
        control_plane_outdated_locations=set(),
    )

    # dep1's error/outdated state was replaced: dep1 location1 now has error, no outdated
    # dep2's state was preserved: dep2 location1 still has error from before
    assert user_code_launcher._control_plane_error_locations == {
        ("dep1", "location1"),
        ("dep2", "location1"),
    }
    # dep1's outdated was cleared (we passed empty set for dep1), dep2 had no outdated
    assert user_code_launcher._control_plane_outdated_locations == set()

    # Now add upload metadata for dep2 — dep1's state must be retained
    user_code_launcher.add_upload_metadata_for_deployment(
        "dep2",
        upload_metadata={("dep2", "location1"): dep2_loc},
        control_plane_error_locations=set(),
        control_plane_outdated_locations={("dep2", "location1")},
    )

    # dep1's error state preserved, dep2's error cleared
    assert user_code_launcher._control_plane_error_locations == {("dep1", "location1")}
    # dep2's outdated set, dep1 had no outdated
    assert user_code_launcher._control_plane_outdated_locations == {("dep2", "location1")}


def test_graphql_uploads(
    caplog: pytest.LogCaptureFixture, user_code_launcher: UserCodeTestLauncher
):
    user_code_launcher.start()

    now = time.time()

    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )

    metadata_two = CodeLocationDeployData("test_image:tag2", python_file="foo.py")

    server_two = UserCodeLauncherEntry(metadata_two, now)

    _check_update(
        user_code_launcher, {("dep1", "location1"): server_one, ("dep1", "location2"): server_two}
    )

    server_two_new_timestamp = UserCodeLauncherEntry(metadata_two, now + 1)

    server_three = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag3", python_file="foo.py"), now
    )

    # Tell the launcher to upload location1 (no changes), location2 (same metadata, new timestamp),
    # and location3 (brand new location). All three should be uploaded as part of the
    # reconciliation

    caplog.clear()

    metadata_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_one,
        ("dep1", "location2"): server_two_new_timestamp,
        ("dep1", "location3"): server_three,
    }
    user_code_launcher.add_upload_metadata_for_deployment(
        "dep1",
        metadata_map,
        control_plane_error_locations=set(),
        control_plane_outdated_locations=set(metadata_map.keys()),
    )

    user_code_launcher.wait_for_reconcile()

    user_code_launcher.assert_reconciliation(metadata_map)

    assert sorted(user_code_launcher.uploaded_locations) == sorted(
        ["location1", "location2", "location3"]
    )

    assert (
        caplog.text.count(
            f"Reconciling to reach {{(dep1, location1, {now}), (dep1, location2, {now + 1}), (dep1,"
            f" location3, {now})}}. To add: {{(dep1, location3, {now})}}. To update: {{(dep1,"
            f" location2, {now + 1})}}. To remove: {{}}. To upload: {{(dep1, location1, {now}),"
            f" (dep1, location2, {now + 1}), (dep1, location3, {now})}}."
        )
        == 1
    )

    for location in ["location1", "location2", "location3"]:
        assert caplog.text.count(f"Fetching metadata for dep1:{location}") == 1

    # only uploads once

    caplog.clear()
    user_code_launcher.wait_for_reconcile()
    assert caplog.text.count("Fetching metadata for") == 0

    assert sorted(user_code_launcher.uploaded_locations) == sorted(
        ["location1", "location2", "location3"]
    )


def test_launcher_invalid_images(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start()

    now = time.time()

    server_with_invalid_image = UserCodeLauncherEntry(
        CodeLocationDeployData(python_file="foo.py", image="trailing_whitespace "),
        now,
    )
    expected_errors = {("dep1", "location1"): "Images must not have leading or trailing whitespace"}

    _check_update(
        user_code_launcher,
        {("dep1", "location1"): server_with_invalid_image},
        expected_entry_map={},
        agent_error_map=expected_errors,
    )

    server_with_invalid_image = UserCodeLauncherEntry(
        CodeLocationDeployData(python_file="foo.py", image="  leading_whitespace"),
        now,
    )

    _check_update(
        user_code_launcher,
        {("dep1", "location1"): server_with_invalid_image},
        expected_entry_map={},
        agent_error_map=expected_errors,
    )


def test_launcher_requires_images(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start()
    assert user_code_launcher.requires_images

    now = time.time()

    server_without_image = UserCodeLauncherEntry(CodeLocationDeployData(python_file="foo.py"), now)
    expected_errors = {
        ("dep1", "location1"): "Your agent's configuration requires you to specify an image."
    }

    _check_update(
        user_code_launcher,
        {("dep1", "location1"): server_without_image},
        expected_entry_map={},
        agent_error_map=expected_errors,
    )

    # updating to a server with image works though
    server_with_image = UserCodeLauncherEntry(
        CodeLocationDeployData(python_file="foo.py", image="now_with_image"), now + 1
    )
    _check_update(user_code_launcher, {("dep1", "location1"): server_with_image})

    # updating back from image to no image fails
    _check_update(
        user_code_launcher,
        {("dep1", "location1"): server_without_image._replace(update_timestamp=now + 2)},
        expected_entry_map={},
        agent_error_map=expected_errors,
    )


def test_launcher_without_images(user_code_launcher_no_images):
    user_code_launcher_no_images.start()
    assert not user_code_launcher_no_images.requires_images

    now = time.time()

    server_without_image = UserCodeLauncherEntry(CodeLocationDeployData(python_file="foo.py"), now)
    _check_update(user_code_launcher_no_images, {("dep1", "location1"): server_without_image})

    updated_server_without_image = UserCodeLauncherEntry(
        CodeLocationDeployData(python_file="bar.py"), now + 1
    )
    _check_update(
        user_code_launcher_no_images, {("dep1", "location1"): updated_server_without_image}
    )

    server_with_image = UserCodeLauncherEntry(
        CodeLocationDeployData(python_file="foo.py", image="now_with_image"), now + 2
    )
    # updating to a server with an image fails

    error_message = "Your agent's configuration cannot load locations that specify a Docker image."

    _check_update(
        user_code_launcher_no_images,
        {("dep1", "location1"): server_with_image},
        expected_entry_map={},
        agent_error_map={("dep1", "location1"): error_message},
    )

    updated_server_without_image = updated_server_without_image._replace(update_timestamp=now + 3)

    # updating back to a server without an image works
    _check_update(
        user_code_launcher_no_images, {("dep1", "location1"): updated_server_without_image}
    )

    server_with_image = server_with_image._replace(update_timestamp=now + 4)

    # can't add a server with an image
    _check_update(
        user_code_launcher_no_images,
        {
            ("dep1", "location1"): updated_server_without_image,
            ("dep1", "location2"): server_with_image,
        },
        expected_entry_map={("dep1", "location1"): updated_server_without_image},
        agent_error_map={("dep1", "location2"): error_message},
    )


def test_cleanup_after_add_failure(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start()

    now = time.time()

    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )

    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="foo.py"), now + 1
    )

    _check_update(user_code_launcher, {("dep1", "location1"): server_one})

    assert len(user_code_launcher.servers[("dep1", "location1")]) == 1

    user_code_launcher.fail_next_add_after_server_added = {"location1"}

    _check_update(
        user_code_launcher,
        {("dep1", "location1"): server_two},
        expected_entry_map={},
        agent_error_map={("dep1", "location1"): "Server add failed after creating the server"},
    )

    # Old server is cleaned up after the failure, new server stays in an error state
    assert len(user_code_launcher.servers[("dep1", "location1")]) == 1


def test_truncate_uploaded_entries(user_code_launcher: UserCodeTestLauncher):
    # Verify that the error is truncated
    user_code_launcher.start()

    now = time.time()
    server = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="foo.py"), now
    )
    with environ({"DAGSTER_CLOUD_CODE_LOCATION_UPLOAD_ERROR_SIZE_LIMIT": "25"}):
        metadata_map = {("dep1", "location1"): server}
        user_code_launcher.add_upload_metadata_for_deployment(
            "dep1",
            upload_metadata=metadata_map,
            control_plane_error_locations=set(),
            control_plane_outdated_locations=set(metadata_map.keys()),
        )

        user_code_launcher.fail_next_add_after_server_added = {"location1"}

        _check_update(
            user_code_launcher,
            metadata_map,
            expected_entry_map={},
            agent_error_map={("dep1", "location1"): "Server add failed after creating the server"},
        )

        assert len(user_code_launcher.uploaded_entries) == 1
        assert user_code_launcher.uploaded_entries[0].serialized_error_info

        assert (
            "Exception: Server add fai (TRUNCATED)"
            in user_code_launcher.uploaded_entries[0].serialized_error_info.message
        )


def test_failure_recovery(
    caplog: pytest.LogCaptureFixture, user_code_launcher: UserCodeTestLauncher
):
    user_code_launcher.start()

    now = time.time()

    # Add a server
    server_one = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag1", python_file="foo.py"), now
    )
    _check_update(user_code_launcher, {("dep1", "location1"): server_one})

    server_two = UserCodeLauncherEntry(
        CodeLocationDeployData("test_image:tag2", python_file="bar.py"), now
    )

    metadata_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_one,
        ("dep1", "location2"): server_two,
    }

    # Test recovery after adding a server
    user_code_launcher.fail_next_add = True

    # location2 is new — server doesn't have data for it yet
    user_code_launcher.update_grpc_metadata(
        metadata_map,
        control_plane_error_locations=set(),
        control_plane_outdated_locations={("dep1", "location2")},
    )
    user_code_launcher.wait_for_reconcile()

    # Immediately update control_plane_error_locations to reflect that the server now knows
    # about location2's error (agent uploaded it). This prevents the recovery mechanism
    # from retrying on the next reconcile cycle.
    location2_error = {("dep1", "location2")}
    user_code_launcher.update_grpc_metadata(
        metadata_map,
        control_plane_error_locations=location2_error,
        control_plane_outdated_locations=set(),
    )

    assert caplog.text.count("Error while updating server for dep1:location2") == 1

    expected_metadata_map: UserCodeLauncherEntryMap = {("dep1", "location1"): server_one}
    expected_errors = {("dep1", "location2"): "Server add failed"}

    # Don't wait for another reconcile - just check current state
    user_code_launcher.assert_reconciliation(
        expected_metadata_map, expected_errors, run_reconcile_thread=False
    )

    # Retry keeps in the same state since metadata did not change.
    _check_update(
        user_code_launcher,
        metadata_map,
        expected_metadata_map,
        expected_errors,
        control_plane_error_locations=location2_error,
    )

    # But changing the metadata causes the location to reload again
    server_three = UserCodeLauncherEntry(
        CodeLocationDeployData("test_another_image:tag3", python_file="bar.py"), now + 1
    )

    metadata_map: UserCodeLauncherEntryMap = {
        ("dep1", "location1"): server_one,
        ("dep1", "location2"): server_three,
    }
    _check_update(user_code_launcher, metadata_map, control_plane_error_locations=location2_error)

    # Test recovery after updating server metadata if add fails
    server_four = UserCodeLauncherEntry(
        CodeLocationDeployData("test_another_image:tag3", python_file="bar.py"), now + 2
    )
    metadata_map = {("dep1", "location1"): server_four, ("dep1", "location2"): server_three}

    user_code_launcher.fail_next_add = True
    # location1 has new timestamp — server data is stale
    user_code_launcher.update_grpc_metadata(
        metadata_map,
        control_plane_error_locations=set(),
        control_plane_outdated_locations={("dep1", "location1")},
    )
    user_code_launcher.wait_for_reconcile()

    # Immediately update control_plane_error_locations to reflect that the server now knows
    # about location1's error (agent uploaded it).
    location1_error = {("dep1", "location1")}
    user_code_launcher.update_grpc_metadata(
        metadata_map,
        control_plane_error_locations=location1_error,
        control_plane_outdated_locations=set(),
    )

    expected_metadata_map = {("dep1", "location2"): server_three}
    expected_errors = {("dep1", "location1"): "Server add failed"}

    # Don't wait for another reconcile - just check current state
    user_code_launcher.assert_reconciliation(
        expected_metadata_map, expected_errors, run_reconcile_thread=False
    )

    # Retry keeps in the same state since timestamp did not change.
    _check_update(
        user_code_launcher,
        metadata_map,
        expected_metadata_map,
        expected_errors,
        control_plane_error_locations=location1_error,
    )

    # But changing the timestamp causes the location to reload again
    server_five = UserCodeLauncherEntry(
        CodeLocationDeployData("test_another_image:tag3", python_file="bar.py"), now + 3
    )
    metadata_map = {("dep1", "location1"): server_five, ("dep1", "location2"): server_three}

    _check_update(user_code_launcher, metadata_map, control_plane_error_locations=location1_error)

    caplog.clear()

    # Test recovery after updating server metadata if remove fails
    metadata_map = {("dep1", "location1"): server_five, ("dep1", "location2"): server_two}
    user_code_launcher.fail_next_remove = True

    # Despite the error, metadata map is unchanged (but there is an error logged)
    _check_update(user_code_launcher, metadata_map)

    assert (
        caplog.text.count("Error while cleaning up after updating server for dep1:location2") == 1
    )

    caplog.clear()

    # Retry keeps in the same state since metadata did not change
    _check_update(user_code_launcher, metadata_map)

    assert (
        caplog.text.count("Error while cleaning up after updating server for dep1:location2") == 0
    )

    # But changing the timestamp causes the location to reload again
    metadata_map = {("dep1", "location1"): server_five, ("dep1", "location2"): server_three}
    _check_update(user_code_launcher, metadata_map)

    # Test recovery after deleting server metadata
    metadata_map = {("dep1", "location2"): server_three}
    user_code_launcher.fail_next_remove = True

    caplog.clear()

    user_code_launcher.update_grpc_metadata(
        metadata_map, control_plane_error_locations=set(), control_plane_outdated_locations=set()
    )
    user_code_launcher.wait_for_reconcile()

    # Despite the error, metadata map is unchanged (but an error is logged)
    user_code_launcher.assert_reconciliation(metadata_map)

    assert caplog.text.count("Error while removing server for dep1:location1") == 1

    # Retry updates to expected state
    _check_update(user_code_launcher, metadata_map)


def test_can_register_instance(user_code_launcher: UserCodeTestLauncher):
    assert isinstance(
        # We are accessing a private property in this test, but
        # in practice this property is always accessed from within
        # the class itself.
        user_code_launcher._instance,
        DagsterInstance,
    )


def test_dont_upload_on_start(user_code_launcher, monkeypatch):
    mock_gql_response = {
        "data": {
            "workspace": {
                "workspaceEntries": [
                    {"locationName": "test-location"},
                ],
            },
        },
    }
    monkeypatch.setattr(
        "dagster_cloud_cli.core.graphql_client.DagsterCloudGraphQLClient.execute",
        lambda *args, **kwargs: mock_gql_response,
    )

    # Typically, we don't upload on start
    user_code_launcher.start()
    assert not user_code_launcher._upload_locations


def test_pex_metadata(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start()

    now = time.time()

    # Add a pex location (which should create a multipex server and a grpc server subprocess on it)
    pex_server_one = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_one})
    assert len(user_code_launcher.multipex_server_handles) == 1

    multipex_server = user_code_launcher.multipex_server_handles[("dep1", "location1")]

    pex_servers = user_code_launcher._get_existing_pex_servers("dep1", "location1").copy()
    assert len(pex_servers) == 1

    # operation is idempotent
    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_one})
    assert len(user_code_launcher.multipex_server_handles) == 1
    assert user_code_launcher._get_existing_pex_servers("dep1", "location1") == pex_servers

    assert user_code_launcher.multipex_server_handles[("dep1", "location1")] == multipex_server

    new_now = now + 10.5

    # update to a new location with the same base image (same multipex server, new grpc server)
    pex_server_two = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        new_now,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_two})

    # Multipex server unchanged
    assert user_code_launcher.multipex_server_handles[("dep1", "location1")] == multipex_server

    # Pex server changed
    new_pex_servers = user_code_launcher._get_existing_pex_servers("dep1", "location1")
    assert len(new_pex_servers) == 1
    assert new_pex_servers != pex_servers

    new_now = now + 20.5

    # Changing the base image, though, does swap out the multipex server
    pex_server_three = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag2",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        new_now,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_three})
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 1
    assert user_code_launcher.multipex_server_handles[("dep1", "location1")] != multipex_server

    assert user_code_launcher._get_existing_pex_servers("dep1", "location1")

    # Using a python version instead of an image tag changes out the multipex server
    multipex_server = user_code_launcher.multipex_server_handles[("dep1", "location1")]
    new_now = new_now + 1

    pex_server_four = UserCodeLauncherEntry(
        CodeLocationDeployData(
            None,
            python_file="foo.py",
            pex_metadata=PexMetadata(
                pex_tag="files=deps-hash.pex:source-hash.pex", python_version="3.9"
            ),
        ),
        new_now,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_four})
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 1
    assert user_code_launcher.multipex_server_handles[("dep1", "location1")] != multipex_server

    assert user_code_launcher._get_existing_pex_servers("dep1", "location1")

    # Same pex tag with new python version changes the multipex server
    multipex_server = user_code_launcher.multipex_server_handles[("dep1", "location1")]
    new_now = new_now + 1

    pex_server_five = UserCodeLauncherEntry(
        CodeLocationDeployData(
            None,
            python_file="foo.py",
            pex_metadata=PexMetadata(
                pex_tag="files=deps-hash.pex:source-hash.pex", python_version="3.12"
            ),
        ),
        new_now,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_five})
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 1
    assert user_code_launcher.multipex_server_handles[("dep1", "location1")] != multipex_server
    assert user_code_launcher._get_existing_pex_servers("dep1", "location1")

    # Same python version with a new pex tag reuses the multipex server
    multipex_server = user_code_launcher.multipex_server_handles[("dep1", "location1")]
    new_now = new_now + 40

    pex_server_six = UserCodeLauncherEntry(
        CodeLocationDeployData(
            None,
            python_file="foo.py",
            pex_metadata=PexMetadata(
                pex_tag="files=deps-hash2.pex:source-hash.pex", python_version="3.12"
            ),
        ),
        new_now,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_six})
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 1
    assert user_code_launcher.multipex_server_handles[("dep1", "location1")] == multipex_server
    assert user_code_launcher._get_existing_pex_servers("dep1", "location1")

    new_now = new_now + 20.5

    # Add a second location

    _check_update(
        user_code_launcher,
        {("dep1", "location1"): pex_server_three, ("dep1", "location2"): pex_server_two},
    )

    assert len(user_code_launcher.multipex_server_handles) == 2

    # Remove a server

    assert ("dep1", "location1") in user_code_launcher._multipex_servers
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location2")]) == 1
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 1

    _check_update(user_code_launcher, {("dep1", "location2"): pex_server_two})
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location2")]) == 1
    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 0

    assert ("dep1", "location1") not in user_code_launcher._multipex_servers

    # Simulate the case where the loop decides that the pex server has become unavailable adn
    # redeploys it

    old_multipex_handles = user_code_launcher.multipex_server_handles[("dep1", "location2")]

    del user_code_launcher._actual_entries[("dep1", "location2")]
    del user_code_launcher._multipex_servers[("dep1", "location2")]

    _check_update(user_code_launcher, {("dep1", "location2"): pex_server_two})

    # new multipex server spins up

    assert len(user_code_launcher.multipex_server_handles[("dep1", "location2")]) == 1
    assert user_code_launcher.multipex_server_handles[("dep1", "location2")] != old_multipex_handles


def test_pex_failure_recovery_on_add(
    caplog: pytest.LogCaptureFixture, user_code_launcher: UserCodeTestLauncher
):
    # Issue spinning up a multipex server recovers, returns a clear error

    user_code_launcher.start()

    now = time.time()

    pex_server_one = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now,
    )

    user_code_launcher.fail_next_add = True

    metadata_map = {("dep1", "location1"): pex_server_one}
    # location1 is new — server doesn't have data for it yet
    user_code_launcher.update_grpc_metadata(
        metadata_map,
        control_plane_error_locations=set(),
        control_plane_outdated_locations={("dep1", "location1")},
    )
    user_code_launcher.wait_for_reconcile()

    # Immediately update control_plane_error_locations to prevent recovery on the next cycle.
    location1_error = {("dep1", "location1")}
    user_code_launcher.update_grpc_metadata(
        metadata_map,
        control_plane_error_locations=location1_error,
        control_plane_outdated_locations=set(),
    )

    assert caplog.text.count("Error while setting up multipex server for") == 1

    expected_errors = {("dep1", "location1"): "Server add failed"}

    user_code_launcher.assert_reconciliation({}, expected_errors, run_reconcile_thread=False)

    # Recovers when you redeploy (by increasing tiemstamp)
    pex_server_two = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now + 10,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_two})

    multipex_server = user_code_launcher._multipex_servers[("dep1", "location1")]

    # Server becomes unavailable - new multipex server should spin up

    user_code_launcher.fail_next_multipex_health_check = True

    pex_server_three = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now + 20,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_three})

    assert (
        caplog.text.count(
            "Spinning up a new multipex server for dep1:location1 since the existing one failed"
        )
        == 1
    )

    # New server does in fact exist and is different than the old one
    new_multipex_server = user_code_launcher._multipex_servers[("dep1", "location1")]
    assert new_multipex_server != multipex_server

    # idempotent afterwards

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_three})
    assert user_code_launcher._multipex_servers[("dep1", "location1")] == new_multipex_server


def test_pex_failure_waiting_for_new_server(
    caplog: pytest.LogCaptureFixture, user_code_launcher: UserCodeTestLauncher
):
    user_code_launcher.start()

    now = time.time()

    pex_server_one = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now,
    )

    user_code_launcher.fail_next_add_after_server_added = {"location1"}

    _check_update(
        user_code_launcher,
        {("dep1", "location1"): pex_server_one},
        expected_entry_map={},
        agent_error_map={("dep1", "location1"): "Server add failed after creating the server"},
    )

    assert caplog.text.count("Error while waiting for multipex server for dep1:location1") == 1

    # Recovers when you redeploy (by increasing tiemstamp)
    pex_server_two = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now + 10,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_two})


def test_change_from_multipex_to_pex(user_code_launcher: UserCodeTestLauncher):
    user_code_launcher.start()

    now = time.time()

    pex_server_one = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
            pex_metadata=PexMetadata(pex_tag="files=deps-hash.pex:source-hash.pex"),
        ),
        now,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): pex_server_one})

    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 1

    non_pex_server = UserCodeLauncherEntry(
        CodeLocationDeployData(
            "test_image:tag1",
            python_file="foo.py",
        ),
        now + 10,
    )

    _check_update(user_code_launcher, {("dep1", "location1"): non_pex_server})

    assert len(user_code_launcher.multipex_server_handles[("dep1", "location1")]) == 0

    assert not user_code_launcher._multipex_servers.get(("dep1", "location1"))



import logging
import threading
from unittest import mock

from dagster import DagsterRun, DagsterRunStatus
from dagster._core.launcher import CheckRunHealthResult, WorkerStatus
from dagster_cloud.execution.cloud_run_launcher.process import CloudProcessRunLauncher
from dagster_cloud.execution.monitoring import (
    CloudRunWorkerStatus,
    run_worker_monitoring_thread_iteration,
)
from dagster_cloud.instance import DagsterCloudAgentInstance
from dagster_cloud.util import NON_ISOLATED_RUN_TAG_PAIR, SERVER_HANDLE_TAG


def test_cloud_run_worker_status_from_oss_worker_status():
    status = CloudRunWorkerStatus.from_check_run_health_result(
        "my-run", CheckRunHealthResult(WorkerStatus.RUNNING, "my-msg")
    )
    assert status.run_id == "my-run"
    assert status.status_type == WorkerStatus.RUNNING
    assert status.message == "my-msg"


def test_cloud_run_worker_status_with_run_worker_debug_info():
    status = CloudRunWorkerStatus.from_check_run_health_result(
        "my-run", CheckRunHealthResult(WorkerStatus.RUNNING, "my-msg"), "foobar"
    )
    assert status.run_id == "my-run"
    assert status.status_type == WorkerStatus.RUNNING
    assert status.message == "my-msg\nfoobar"


def test_cloud_run_worker_status_with_very_long_message():
    not_so_silent_scream = "".join(["A" for i in range(2000)])

    status = CloudRunWorkerStatus.from_check_run_health_result(
        "my-run",
        CheckRunHealthResult(WorkerStatus.RUNNING, not_so_silent_scream),
        not_so_silent_scream,
    )
    assert status.run_id == "my-run"
    assert status.status_type == WorkerStatus.RUNNING
    assert status.message == "".join(["A" for i in range(1000)])


def test_thread_error(agent_instance, caplog):
    with mock.patch(
        "dagster_cloud.execution.monitoring.get_cloud_run_worker_statuses"
    ) as mock_get_statuses:
        mock_get_statuses.side_effect = Exception("Failure fetching statuses")

        deployments_to_check = {"foo"}
        statuses_dict = {}
        run_worker_monitoring_lock = threading.Lock()
        logger = logging.getLogger("dagster_cloud")

        run_worker_monitoring_thread_iteration(
            agent_instance, deployments_to_check, statuses_dict, run_worker_monitoring_lock, logger
        )

        assert "Failure fetching statuses" in caplog.text


def test_monitoring(
    agent_instance,
):
    with mock.patch.object(DagsterCloudAgentInstance, "get_runs") as mock_get_runs:
        mock_get_runs.return_value = [
            DagsterRun(
                job_name="foo",
                run_id="bar",
                run_config={},
                step_keys_to_execute=None,
                status=DagsterRunStatus.STARTED,
                tags={
                    NON_ISOLATED_RUN_TAG_PAIR[0]: NON_ISOLATED_RUN_TAG_PAIR[1],
                    SERVER_HANDLE_TAG: "non-existent-server",
                },
            )
        ]
        deployments_to_check = {"foo"}
        statuses_dict = {}
        run_worker_monitoring_lock = threading.Lock()
        logger = logging.getLogger("dagster_cloud")

        run_worker_monitoring_thread_iteration(
            agent_instance, deployments_to_check, statuses_dict, run_worker_monitoring_lock, logger
        )

        assert len(statuses_dict["foo"]) == 1
        assert statuses_dict["foo"][0].run_id == "bar"
        assert statuses_dict["foo"][0].status_type == WorkerStatus.FAILED


def test_monitoring_k8s_pod_debug_info(
    agent_instance,
):
    with mock.patch.object(DagsterCloudAgentInstance, "get_runs") as mock_get_runs:
        mock_get_runs.return_value = [
            DagsterRun(
                job_name="foo",
                run_id="bar",
                run_config={},
                step_keys_to_execute=None,
                status=DagsterRunStatus.STARTED,
                tags={},
            )
        ]
        deployments_to_check = {"foo"}
        statuses_dict = {}
        run_worker_monitoring_lock = threading.Lock()
        logger = logging.getLogger("dagster_cloud")

        with mock.patch.object(CloudProcessRunLauncher, "check_run_worker_health") as mock_check:
            mock_check.return_value = CheckRunHealthResult(WorkerStatus.FAILED, "my-msg")

            with mock.patch.object(
                CloudProcessRunLauncher, "get_run_worker_debug_info"
            ) as mock_debug:
                mock_debug.return_value = "my-debug-info"

                run_worker_monitoring_thread_iteration(
                    agent_instance,
                    deployments_to_check,
                    statuses_dict,
                    run_worker_monitoring_lock,
                    logger,
                )

        assert len(statuses_dict["foo"]) == 1
        assert statuses_dict["foo"][0].run_id == "bar"
        assert statuses_dict["foo"][0].status_type == WorkerStatus.FAILED
        assert statuses_dict["foo"][0].message == "my-msg\nmy-debug-info"



# ruff: noqa: SLF001

import datetime
import time
from typing import TYPE_CHECKING
from unittest import mock

import pytest
from dagster._core.test_utils import environ
from dagster_cloud import DagsterCloudAgentInstance
from dagster_cloud.workspace.user_code_launcher import UserCodeLauncherEntry
from dagster_cloud.workspace.user_code_launcher.user_code_launcher import (
    CLEANUP_SERVER_GRACE_PERIOD_SECONDS,
)
from dagster_cloud_cli.core.workspace import CodeLocationDeployData
from freezegun import freeze_time

from dagster_cloud_tests.test_user_code_launcher import UserCodeTestLauncher

if TYPE_CHECKING:
    from dagster_cloud.workspace.user_code_launcher.user_code_launcher import DeploymentAndLocation


class GracefulUserCodeTestLauncher(UserCodeTestLauncher):
    @property
    def supports_get_current_runs_for_server_handle(self) -> bool:
        return True

    def run_launcher(self):
        raise NotImplementedError()

    @property
    def _reconcile_interval(self):  # pyright: ignore[reportIncompatibleMethodOverride]
        return 1


@pytest.fixture
def patched_get_current_runs():
    with mock.patch.object(
        GracefulUserCodeTestLauncher, "get_current_runs_for_server_handle"
    ) as method:
        method.return_value = []
        yield method


@pytest.fixture
def patched_list_server_handles():
    with mock.patch.object(GracefulUserCodeTestLauncher, "_list_server_handles") as method:
        method.return_value = []
        yield method


@pytest.fixture
def user_code_launcher(patched_get_current_runs, patched_list_server_handles, agent_instance):
    with GracefulUserCodeTestLauncher() as user_code_launcher:
        user_code_launcher.register_instance(agent_instance)
        yield user_code_launcher


def test_graceful_multiple_agents_grpc_spindown(agent_instance: DagsterCloudAgentInstance):
    with GracefulUserCodeTestLauncher() as user_code_launcher:
        with environ({"DAGSTER_CLOUD_CLEANUP_SERVER_CHECK_INTERVAL": "0"}):
            active_agent_ids = {"different-agent-id", agent_instance.instance_uuid}
            user_code_launcher._mocked_active_agent_ids = active_agent_ids
            user_code_launcher.register_instance(agent_instance)
            user_code_launcher.start(run_reconcile_thread=False)
            now = time.time()
            one_hour_ago = now - CLEANUP_SERVER_GRACE_PERIOD_SECONDS - 1

            server_one = UserCodeLauncherEntry(
                CodeLocationDeployData("test_image:tag1", python_file="foo.py"),
                now,
            )
            server_two = UserCodeLauncherEntry(
                CodeLocationDeployData("test_image:tag2", python_file="foo.py"), now
            )
            server_three = UserCodeLauncherEntry(
                CodeLocationDeployData("test_image:tag3", python_file="foo.py"), now
            )
            server_four = UserCodeLauncherEntry(
                CodeLocationDeployData("test_image:tag4", python_file="foo.py"), now
            )

            with freeze_time(
                datetime.datetime.fromtimestamp(one_hour_ago, tz=datetime.timezone.utc)
            ):
                entry_map: dict[DeploymentAndLocation, UserCodeLauncherEntry] = {
                    ("dep1", "location1"): server_one,
                    ("dep2", "location2"): server_two,
                    ("dep3", "location3"): server_three,
                }

                user_code_launcher.update_grpc_metadata(
                    entry_map,
                    control_plane_error_locations=set(),
                    control_plane_outdated_locations=set(),
                )
                user_code_launcher.get_grpc_server_heartbeats()
                user_code_launcher.reconcile()
                user_code_launcher.get_grpc_server_heartbeats()

                assert len(user_code_launcher._agent_ids.keys()) == 3
                # Since all servers were created by this agent instance, both should be able to clean up.
                for key in user_code_launcher._agent_ids.keys():
                    assert user_code_launcher._can_cleanup_server(
                        key, active_agent_ids, include_own_servers=True
                    )

            with freeze_time(datetime.datetime.fromtimestamp(now, tz=datetime.timezone.utc)):
                entry_map[("dep4", "location4")] = server_four

                user_code_launcher.update_grpc_metadata(
                    entry_map,
                    control_plane_error_locations=set(),
                    control_plane_outdated_locations=set(),
                )
                user_code_launcher.get_grpc_server_heartbeats()
                user_code_launcher.reconcile()
                user_code_launcher.get_grpc_server_heartbeats()

                assert len(user_code_launcher._agent_ids.keys()) == 4
                # Since all servers were created by this agent instance, both should be able to clean up.
                for key in user_code_launcher._agent_ids.keys():
                    assert user_code_launcher._can_cleanup_server(
                        key, active_agent_ids, include_own_servers=True
                    )

                handle_active_agent = [
                    key
                    for key in user_code_launcher._agent_ids.keys()
                    if key.startswith("location1-dep1")
                ].pop()
                handle_inactive_agent_not_recent = [
                    key
                    for key in user_code_launcher._agent_ids.keys()
                    if key.startswith("location2-dep2")
                ].pop()

                handle_inactive_agent_recent = [
                    key
                    for key in user_code_launcher._agent_ids.keys()
                    if key.startswith("location4-dep4")
                ].pop()

                # Induce a state where one of the grpc servers was created by a different agent which is active.
                user_code_launcher._agent_ids[handle_active_agent] = "different-agent-id"

                assert not user_code_launcher._can_cleanup_server(
                    handle_active_agent, active_agent_ids, include_own_servers=True
                )

                # Induce a state where one of the grpc servers was created by a different agent which is inactive.
                user_code_launcher._agent_ids[handle_inactive_agent_not_recent] = (
                    "inactive-agent-id"
                )
                user_code_launcher._agent_ids[handle_inactive_agent_recent] = "inactive-agent-id"

                # recent ones still cannot be cleaned up
                assert not user_code_launcher._can_cleanup_server(
                    handle_inactive_agent_recent, active_agent_ids, include_own_servers=True
                )

                # non-recent ones can
                assert user_code_launcher._can_cleanup_server(
                    handle_inactive_agent_not_recent, active_agent_ids, include_own_servers=True
                )

                # Perform cleanup of grpc servers
                user_code_launcher._graceful_cleanup_servers(include_own_servers=True)
                remaining_handles = set()
                for handles in user_code_launcher._servers.values():
                    for handle in handles:
                        remaining_handles.add(handle)
                # Ensure that the handle created by this agent, and the handle created by
                # an inactive agent were both cleaned up.
                assert remaining_handles == {handle_active_agent, handle_inactive_agent_recent}


def test_mocks(user_code_launcher, patched_list_server_handles, patched_get_current_runs):
    patched_list_server_handles.return_value = [123]
    patched_get_current_runs.return_value = ["foo"]

    assert user_code_launcher._list_server_handles() == [123]
    assert user_code_launcher.get_current_runs_for_server_handle(123) == ["foo"]


def test_empty(user_code_launcher):
    user_code_launcher.start()
    user_code_launcher.wait_for_reconcile()
    with user_code_launcher._grpc_servers_lock:
        assert user_code_launcher._pending_delete_grpc_server_handles == set()


def test_cleans_up_no_active_runs(
    user_code_launcher, patched_list_server_handles, patched_get_current_runs
):
    patched_list_server_handles.return_value = [123]
    assert user_code_launcher.get_current_runs_for_server_handle(123) == []

    user_code_launcher.start()
    user_code_launcher.wait_for_reconcile()
    with user_code_launcher._grpc_servers_lock:
        assert user_code_launcher._pending_delete_grpc_server_handles == set()


def test_leaves_active_runs(
    user_code_launcher, patched_list_server_handles, patched_get_current_runs
):
    patched_list_server_handles.return_value = [123]
    patched_get_current_runs.return_value = ["f00"]

    user_code_launcher.start()
    user_code_launcher.wait_for_reconcile()
    with user_code_launcher._grpc_servers_lock:
        assert user_code_launcher._pending_delete_grpc_server_handles == {123}
    assert user_code_launcher.get_active_grpc_server_handles() == [123]

    patched_get_current_runs.return_value = []

    user_code_launcher.wait_for_reconcile()
    with user_code_launcher._grpc_servers_lock:
        assert user_code_launcher._pending_delete_grpc_server_handles == set()

    assert user_code_launcher.get_active_grpc_server_handles() == []



from dataclasses import replace

import dagster._check as check
from dagster import (
    AssetCheckEvaluation,
    AssetCheckResult,
    AssetExecutionContext,
    AssetKey,
    AssetMaterialization,
    AssetObservation,
    DagsterInvariantViolationError,
    OpExecutionContext,
    Output,
)
from dagster._core.errors import DagsterInvalidPropertyError


def get_current_context_and_asset_key() -> tuple[
    OpExecutionContext | AssetExecutionContext, AssetKey | None
]:
    asset_key = None
    try:
        context = AssetExecutionContext.get()
        if len(context.assets_def.keys_by_output_name.keys()) == 1:
            asset_key = context.asset_key
    except (DagsterInvalidPropertyError, DagsterInvariantViolationError):
        context = OpExecutionContext.get()

    return check.not_none(context), asset_key


def get_asset_key_for_output(
    context: OpExecutionContext | AssetExecutionContext, output_name: str
) -> AssetKey | None:
    asset_key = context.job_def.asset_layer.get_asset_key_for_node_output(
        node_handle=context.op_handle, output_name=output_name
    )
    if asset_key is None:
        return None
    return asset_key


def extract_asset_info_from_event(
    context,
    dagster_event: Output
    | AssetMaterialization
    | AssetObservation
    | AssetCheckResult
    | AssetCheckEvaluation,
    record_observation_usage,
):
    if isinstance(dagster_event, AssetMaterialization):
        return dagster_event.asset_key, dagster_event.partition

    if (
        isinstance(dagster_event, (AssetCheckResult, AssetObservation, AssetCheckEvaluation))
        and record_observation_usage
    ):
        partition = dagster_event.partition if isinstance(dagster_event, AssetObservation) else None
        return dagster_event.asset_key, partition

    if isinstance(dagster_event, (AssetCheckResult, AssetObservation, AssetCheckEvaluation)):
        return None, None

    if isinstance(dagster_event, Output):
        asset_key = get_asset_key_for_output(context, dagster_event.output_name)
        partition_key = None
        if asset_key and context._step_execution_context.has_asset_partitions_for_output(  # noqa: SLF001
            dagster_event.output_name
        ):
            # We associate cost with the first partition key in the case that an output
            # maps to multiple partitions. This is a temporary solution, but partition key
            # is not used in Insights at the moment.
            # TODO: Find a long-term solution for this
            partition_key = next(
                iter(context.asset_partition_keys_for_output(dagster_event.output_name))
            )

        return asset_key, partition_key

    return None, None


def handle_raise_on_error(invocation_arg_name: str, arg_position: int = 1):
    # defer import since dagster_cloud shouldn't depend on dagster_dbt
    # but this should only be used in envs that have dagster_dbt
    from dagster_dbt import DbtCliInvocation

    def raise_decorator(wrapper_fn):
        def inner(*args, **kwargs):
            from_args = len(args) > arg_position
            dbt_cli_invocation = (
                args[arg_position] if from_args else kwargs.get(invocation_arg_name)
            )
            assert isinstance(dbt_cli_invocation, DbtCliInvocation)

            # If the user has set raise_on_error=True, we want to raise the error, but not before we're
            # able to emit the cost information.  This context manager provides a replacement invocation
            # that will not raise, but will handle errors after the context is returned
            raise_on_error = dbt_cli_invocation.raise_on_error

            if not raise_on_error:
                yield from wrapper_fn(*args, **kwargs)
                return

            stubbed_dbt_cli_invocation = replace(dbt_cli_invocation, raise_on_error=False)
            new_args = (
                [
                    *args[:arg_position],
                    stubbed_dbt_cli_invocation,
                    *args[arg_position + 1 :],
                ]
                if from_args
                else args
            )
            new_kwargs = (
                {
                    **kwargs,
                    invocation_arg_name: stubbed_dbt_cli_invocation,
                }
                if not from_args
                else kwargs
            )
            yield from wrapper_fn(*new_args, **new_kwargs)
            error = stubbed_dbt_cli_invocation.get_error()
            if error and raise_on_error:
                raise error

        return inner

    return raise_decorator



import datetime
from unittest.mock import MagicMock, PropertyMock, patch

import pytest
from dagster import define_asset_job
from dagster._core.definitions.asset_checks.asset_check_spec import AssetCheckSeverity
from dagster._core.definitions.asset_checks.asset_checks_definition import AssetChecksDefinition
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.metadata import JsonMetadataValue, MetadataValue, TextMetadataValue
from dagster._core.definitions.source_asset import SourceAsset
from dagster._core.test_utils import freeze_time, instance_for_test
from dagster._time import create_datetime, get_current_datetime
from dagster_cloud.anomaly_detection.defs import (
    DagsterCloudAnomalyDetectionFailed,
    build_anomaly_detection_freshness_checks,
)
from dagster_cloud.anomaly_detection.types import BetaFreshnessAnomalyDetectionParams


@pytest.fixture
def mock_gql_client_execute():
    with patch(
        "dagster_cloud_cli.core.graphql_client.DagsterCloudGraphQLClient.execute"
    ) as mock_client_execute:
        yield mock_client_execute


@pytest.fixture
def mock_agent_params():
    with patch(
        "dagster._core.instance.DagsterInstance.dagit_url", create=True, new_callable=PropertyMock
    ) as mock_dagit_url:
        mock_dagit_url.return_value = "http://localhost:3000"
        with patch(
            "dagster._core.instance.DagsterInstance.dagster_cloud_agent_token",
            create=True,
            new_callable=PropertyMock,
        ) as mock_agent_token:
            mock_agent_token.return_value = "fake_token"
            yield None


@asset(key_prefix="bar")
def foo_asset():
    pass


foo_src = foo_asset.to_source_asset()


@pytest.mark.parametrize(
    "targeted_asset",
    [
        foo_src,
        foo_asset,
    ],
    ids=["source_asset", "asset"],
)
@pytest.mark.parametrize(
    "params",
    [None, BetaFreshnessAnomalyDetectionParams(sensitivity=0.2)],
)
def test_anomaly_detection_freshness_checks_builder(
    mock_gql_client_execute: MagicMock,
    mock_agent_params: None,
    params: BetaFreshnessAnomalyDetectionParams | None,
    targeted_asset: SourceAsset | AssetsDefinition,
) -> None:
    check = build_anomaly_detection_freshness_checks(
        assets=[targeted_asset],
        params=params,
    )[0]
    assert isinstance(check, AssetChecksDefinition)
    assert check.node_def.name.startswith("anomaly_detection_freshness_check_")
    check_job = define_asset_job("check_job", selection=AssetSelection.all_asset_checks())
    defs = Definitions(
        assets=[targeted_asset],
        asset_checks=[check],
        jobs=[check_job],  # We run checks independently of the underlying asset.
    )
    job_def = defs.resolve_job_def(check_job.name)

    with instance_for_test() as instance:
        # Valid output from the gql query
        mock_gql_client_execute.return_value = {
            "data": {
                "anomalyDetectionInference": {
                    "__typename": "AnomalyDetectionSuccess",
                    "response": {
                        "last_updated_timestamp": create_datetime(
                            2020, 3, 5, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "evaluation_timestamp": create_datetime(
                            2020, 3, 5, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "last_update_time_lower_bound": create_datetime(
                            2020, 3, 4, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "maximum_acceptable_delta": 1.0,
                        "model_training_range_start_timestamp": create_datetime(
                            2020, 3, 1, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "model_training_range_end_timestamp": create_datetime(
                            2020, 3, 5, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                    },
                }
            }
        }
        result = job_def.execute_in_process(instance=instance)
        assert result.success
        evaluations = result.get_asset_check_evaluations()
        assert len(evaluations) == 1
        assert evaluations[0].passed
        assert {key: val for key, val in evaluations[0].metadata.items()} == {
            "dagster/anomaly_detection_model_params": JsonMetadataValue(
                {
                    "sensitivity": 0.1 if params is None else params.sensitivity,
                }
            ),
            "dagster/anomaly_detection_model_version": TextMetadataValue("FRESHNESS_BETA"),
            "dagster/anomaly_detection_model_training_range_start_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 1, 0, 0, 0, tz="UTC")
            ),
            "dagster/anomaly_detection_model_training_range_end_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 5, 0, 0, 0, tz="UTC")
            ),
            "dagster/last_updated_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 5, 0, 0, 0, tz="UTC")
            ),
            "dagster/freshness_lower_bound_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 4, 0, 0, 0, tz="UTC")
            ),
            "dagster/fresh_until_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 5, 0, 0, 1, tz="UTC")
            ),
        }
        mock_gql_client_execute.assert_called_once()
        _, gql_params = mock_gql_client_execute.call_args.args
        assert gql_params["modelVersion"] == "FRESHNESS_BETA"
        assert gql_params["params"]["sensitivity"] == 0.1 if params is None else params.sensitivity
        assert gql_params["params"]["asset_key_user_string"] == "bar/foo_asset"

        # Valid overdue output from the gql query
        mock_gql_client_execute.return_value = {
            "data": {
                "anomalyDetectionInference": {
                    "__typename": "AnomalyDetectionSuccess",
                    "response": {
                        "last_updated_timestamp": create_datetime(
                            2020, 3, 1, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "evaluation_timestamp": create_datetime(
                            2020, 3, 1, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "last_update_time_lower_bound": create_datetime(
                            2020, 3, 5, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "maximum_acceptable_delta": 1.0,
                        "model_training_range_start_timestamp": create_datetime(
                            2020, 3, 1, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                        "model_training_range_end_timestamp": create_datetime(
                            2020, 3, 1, 0, 0, 0, tz="UTC"
                        ).timestamp(),
                    },
                }
            }
        }

        result = job_def.execute_in_process(instance=instance)
        assert result.success
        evaluations = result.get_asset_check_evaluations()
        assert len(evaluations) == 1
        assert not evaluations[0].passed
        assert {key: val for key, val in evaluations[0].metadata.items()} == {
            "dagster/anomaly_detection_model_params": JsonMetadataValue(
                {
                    "sensitivity": 0.1 if params is None else params.sensitivity,
                }
            ),
            "dagster/anomaly_detection_model_version": TextMetadataValue("FRESHNESS_BETA"),
            "dagster/anomaly_detection_model_training_range_start_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 1, 0, 0, 0, tz="UTC")
            ),
            "dagster/anomaly_detection_model_training_range_end_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 1, 0, 0, 0, tz="UTC")
            ),
            "dagster/last_updated_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 1, 0, 0, 0, tz="UTC")
            ),
            "dagster/freshness_lower_bound_timestamp": MetadataValue.timestamp(
                create_datetime(2020, 3, 5, 0, 0, 0, tz="UTC")
            ),
        }
        assert evaluations[0].severity == AssetCheckSeverity.WARN

        _, gql_params = mock_gql_client_execute.call_args.args
        assert gql_params["modelVersion"] == "FRESHNESS_BETA"
        assert gql_params["params"]["sensitivity"] == 0.1 if params is None else params.sensitivity
        assert gql_params["params"]["asset_key_user_string"] == "bar/foo_asset"

        # Graphql query failure
        mock_gql_client_execute.return_value = {
            "data": {
                "anomalyDetectionInference": {
                    "__typename": "AnomalyDetectionFailure",
                    "message": "Exception: Anomaly detection failed.",
                }
            }
        }
        with pytest.raises(DagsterCloudAnomalyDetectionFailed, match=r"Anomaly detection failed."):
            job_def.execute_in_process(instance=instance)

        # Not enough records
        mock_gql_client_execute.return_value = {
            "data": {
                "anomalyDetectionInference": {
                    "__typename": "AnomalyDetectionFailure",
                    "message": "Not enough records found to detect anomalies. Need at least 15.",
                }
            }
        }
        freeze_datetime = get_current_datetime()
        with freeze_time(freeze_datetime):
            result = job_def.execute_in_process(instance=instance)
            assert result.success
            evaluations = result.get_asset_check_evaluations()
            assert len(evaluations) == 1
            assert evaluations[0].passed
            assert evaluations[0].severity == AssetCheckSeverity.WARN
            assert (
                evaluations[0].description
                == "Anomaly detection failed: Not enough records found to detect anomalies. Need at "
                "least 15. Any sensors will wait to re-evaluate this check for a day."
            )
            assert evaluations[0].metadata[
                "dagster/fresh_until_timestamp"
            ] == MetadataValue.timestamp(freeze_datetime + datetime.timedelta(days=1))



import json
import os
import sys
import tempfile
import threading
from contextlib import contextmanager
from pathlib import Path

import pytest
import uvicorn
import yaml
from dagster._core.test_utils import environ
from dagster_cloud.auth.constants import get_hardcoded_test_agent_token
from dagster_cloud_cli.entrypoint import app
from dagster_shared.utils import find_free_port
from typer.testing import CliRunner

runner = CliRunner()


@contextmanager
def _run_dagit_server(host_unscoped_instance):
    """Run the dagit server on a real port for subprocess access."""
    from dagster_cloud_backend.workspace import CloudDagitProcessContext
    from dagster_cloud_dagit.server.cloud_dagit import CloudDagitWebserver

    port = find_free_port()

    with environ({"SESSION_SECRET_KEY": "fake"}):
        context = CloudDagitProcessContext(host_unscoped_instance, version="test")
        app = CloudDagitWebserver(context).create_asgi_app(debug=True)

        config = uvicorn.Config(app, host="127.0.0.1", port=port, log_level="warning")
        server = uvicorn.Server(config)

        thread = threading.Thread(target=server.run, daemon=True)
        thread.start()

        # Wait for server to start
        import time

        for _ in range(50):  # Wait up to 5 seconds
            try:
                import httpx

                httpx.get(f"http://127.0.0.1:{port}/server_info", timeout=0.1)
                break
            except Exception:
                time.sleep(0.1)

        try:
            yield f"http://127.0.0.1:{port}"
        finally:
            server.should_exit = True
            thread.join(timeout=5)


@pytest.fixture
def dagit_server_url(host_unscoped_instance):
    """Fixture that provides a real HTTP server URL for the dagit app."""
    with _run_dagit_server(host_unscoped_instance) as url:
        yield url


def _setup_project(project_dir: Path):
    """Adds a dagster_cloud.yaml file and a state-backed component to the project."""
    import shutil

    component_dir = project_dir / "src/foo_bar/defs/the_component"
    component_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy(
        Path(__file__).parent / "state_backed_components" / "sample_state_backed_component.py",
        component_dir / "local.py",
    )
    # Create the defs.yaml file
    with (component_dir / "defs.yaml").open("w") as f:
        yaml.dump(
            {"type": ".local.SampleStateBackedComponent"},
            f,
        )

    # Create the dagster_cloud.yaml file
    with open(project_dir.joinpath("dagster_cloud.yaml"), "w") as f:
        dagster_cloud_config = {
            "locations": [
                {
                    "location_name": "test-location",
                    "image": "test-image:latest",
                    "code_source": {"python_file": "test_file.py"},
                }
            ]
        }

        yaml.dump(dagster_cloud_config, f)


def _update_location_state_url(state_dir: Path, new_url: str):
    """Update the URL in all location state files to point to the test server."""
    for state_file in state_dir.glob("location-*.json"):
        with open(state_file) as f:
            state = json.load(f)
        state["url"] = new_url
        with open(state_file, "w") as f:
            json.dump(state, f)


def test_refresh_defs_state(
    monkeypatch: pytest.MonkeyPatch,
    dagit_server_url: str,
) -> None:
    from dagster_test.dg_utils.utils import ProxyRunner, isolated_example_project_foo_bar

    organization_name = "acme"

    with (
        ProxyRunner.test() as dg_runner,
        isolated_example_project_foo_bar(dg_runner, uv_sync=True) as project_dir,
        tempfile.TemporaryDirectory() as tmp_dir,
    ):
        _setup_project(project_dir)

        sys.path.append(os.fspath(project_dir))

        dagster_cloud_state_dir = Path(tmp_dir) / "build"
        dagster_cloud_state_dir.mkdir()

        monkeypatch.setenv("DAGSTER_CLOUD_ORGANIZATION", organization_name)
        monkeypatch.setenv(
            "DAGSTER_CLOUD_API_TOKEN", get_hardcoded_test_agent_token(organization_name)
        )
        monkeypatch.setenv("DAGSTER_BUILD_STATEDIR", os.fspath(dagster_cloud_state_dir))

        # prod ci run
        monkeypatch.setenv("DAGSTER_CLOUD_DEPLOYMENT", "prod")

        # use dagster-cloud ci init
        result = runner.invoke(
            app,
            [
                "ci",
                "init",
                "--statedir",
                os.fspath(dagster_cloud_state_dir),
                "--project-dir",
                os.fspath(project_dir),
            ],
        )
        assert result.exit_code == 0, result.output

        # Update the location state URL to point to our test server instead of https://acme.dagster.cloud
        # This is needed because the subprocess will make real HTTP requests
        # The URL needs to include the organization prefix since dagit routes are /{org}/graphql
        _update_location_state_url(
            dagster_cloud_state_dir, f"{dagit_server_url}/{organization_name}"
        )

        # now refresh the defs state
        result = dg_runner.invoke(
            "plus",
            "deploy",
            "refresh-defs-state",
        )

        assert result.exit_code == 0, result.output
        assert "Updated defs_state_info for all 1 locations" in result.output, result.output

        sys.path.remove(os.fspath(project_dir))



import unittest
from concurrent.futures import ThreadPoolExecutor
from unittest.mock import patch

from dagster_cloud.batching.batcher import (
    DEFAULT_MAX_BATCH_SIZE,
    DEFAULT_MAX_QUEUE_SIZE,
    DEFAULT_MAX_WAIT_MS,
    Batcher,
)


class BatchingTestCase(unittest.TestCase):
    def test_batching(self):
        batches = _test(list(range(10)))
        assert sum(len(b) for b in batches) == 10

    def test_batching_maxsize(self):
        batches = _test(list(range(20)), max_batch_size=5)
        for batch in batches:
            assert len(batch) <= 5

    def test_batch_failure(self):
        def handler(batch: list[int]) -> list[int]:
            raise Exception("failing")

        batcher = Batcher("test", handler)

        def _do_submit(i: int) -> None:
            with self.assertRaises(Exception, msg="failing"):
                batcher.submit(i)

        with ThreadPoolExecutor() as executor:
            list(executor.map(_do_submit, list(range(10))))

    def test_config(self):
        def handler(f: list[int]) -> list[int]:
            return f

        batcher: Batcher[int, int] = Batcher(
            "test",
            handler,
        )
        assert batcher._max_wait_ms == DEFAULT_MAX_WAIT_MS
        assert batcher._max_batch_size == DEFAULT_MAX_BATCH_SIZE
        assert batcher._queue.maxsize == DEFAULT_MAX_QUEUE_SIZE

        batcher: Batcher[int, int] = Batcher(
            "test",
            handler,
            max_wait_ms=10,
            max_batch_size=11,
            max_queue_size=12,
        )
        assert batcher._max_wait_ms == 10
        assert batcher._max_batch_size == 11
        assert batcher._queue.maxsize == 12

        with patch.dict(
            "os.environ",
            {
                "DAGSTER_BATCHING__TEST__MAX_WAIT_MS": "100",
                "DAGSTER_BATCHING__TEST__MAX_BATCH_SIZE": "101",
                "DAGSTER_BATCHING__TEST__MAX_QUEUE_SIZE": "102",
            },
        ):
            batcher: Batcher[int, int] = Batcher(
                "test",
                handler,
                max_wait_ms=10,
                max_batch_size=11,
                max_queue_size=12,
            )
            assert batcher._max_wait_ms == 100
            assert batcher._max_batch_size == 101
            assert batcher._queue.maxsize == 102


def _test(items: list[int], max_batch_size: int | None = None) -> list[list[int]]:
    batches = []

    def handler(batch: list[int]) -> list[int]:
        batches.append(batch)
        return [x + 1 for x in batch]

    batcher = Batcher("test", handler, max_batch_size=max_batch_size)

    def run():
        def _do_submit(i: int) -> None:
            assert batcher.submit(i) == i + 1

        with ThreadPoolExecutor() as executor:
            list(executor.map(_do_submit, items))

    run()
    return batches



# ruff: noqa: SLF001
import time
from contextlib import contextmanager

import docker
import pytest
from dagster._core.test_utils import instance_for_test
from dagster_cloud.workspace.docker import (
    AGENT_LABEL,
    GRPC_SERVER_LABEL,
    STOP_TIMEOUT_LABEL,
    DockerUserCodeLauncher,
)
from dagster_cloud.workspace.docker.utils import unique_docker_resource_name
from dagster_cloud.workspace.user_code_launcher import DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT
from dagster_cloud.workspace.user_code_launcher.user_code_launcher import UserCodeLauncherEntry
from dagster_cloud.workspace.user_code_launcher.utils import deterministic_label_for_location
from dagster_cloud_cli.core.workspace import CodeLocationDeployData


@pytest.fixture
def docker_client():
    client = docker.client.from_env()

    existing_containers = client.containers.list(all=True)

    yield client

    for container in client.containers.list(all=True):
        if container not in existing_containers:
            container.stop()
            container.remove(force=True)


def test_config():
    assert DockerUserCodeLauncher.config_type()


@contextmanager
def docker_instance(user_code_launcher_overrides=None):
    with instance_for_test(
        {
            "instance_class": {
                "module": "dagster_cloud",
                "class": "DagsterCloudAgentInstance",
            },
            "user_code_launcher": {
                "module": "dagster_cloud.workspace.docker",
                "class": "DockerUserCodeLauncher",
                "config": user_code_launcher_overrides or {},
            },
            "dagster_cloud_api": {
                "url": "http://localhost:2874",
                "agent_token": "FAKE_TOKEN",
            },
            "compute_logs": {
                "module": "dagster.core.storage.noop_compute_log_manager",
                "class": "NoOpComputeLogManager",
            },
        }
    ) as instance:
        yield instance


def test_default_instance():
    with docker_instance() as instance:
        assert instance.user_code_launcher.env_vars == []  # pyright: ignore[reportAttributeAccessIssue]
        assert instance.user_code_launcher.container_kwargs == {}  # pyright: ignore[reportAttributeAccessIssue]
        assert (
            instance.user_code_launcher._server_process_startup_timeout  # pyright: ignore[reportAttributeAccessIssue]
            == DEFAULT_SERVER_PROCESS_STARTUP_TIMEOUT
        )


def test_container_kwargs():
    container_kwargs = {"auto_remove": True}
    with docker_instance({"container_kwargs": container_kwargs}) as instance:
        assert instance.run_launcher.container_kwargs == container_kwargs  # pyright: ignore[reportAttributeAccessIssue]


def test_override_timeout():
    with docker_instance({"server_process_startup_timeout": 1234}) as instance:
        assert instance.user_code_launcher._server_process_startup_timeout == 1234  # pyright: ignore[reportAttributeAccessIssue]


def test_get_standalone_server_handles_for_location(docker_client):
    with docker_instance() as instance:
        assert not instance.user_code_launcher._get_standalone_dagster_server_handles_for_location(  # pyright: ignore[reportAttributeAccessIssue]
            deployment_name="foo",
            location_name="bar",
        )

        docker_client.images.pull("dagster/dagster-cloud-examples:1.12.5")
        docker_client.containers.create(
            image="dagster/dagster-cloud-examples:1.12.5",
            labels={
                GRPC_SERVER_LABEL: "",
                deterministic_label_for_location("foo", "bar"): "",
                AGENT_LABEL: instance.instance_uuid,  # pyright: ignore[reportAttributeAccessIssue]
            },
        )

        handles = instance.user_code_launcher._get_standalone_dagster_server_handles_for_location(  # pyright: ignore[reportAttributeAccessIssue]
            deployment_name="foo",
            location_name="bar",
        )

        assert len(handles) == 1

        handle = handles[0]
        create_timestamp = instance.user_code_launcher.get_server_create_timestamp(handle)  # pyright: ignore[reportAttributeAccessIssue]

        assert create_timestamp <= time.time() and create_timestamp >= time.time() - 60 * 5


def test_long_docker_resource_name():
    long_deployment_name = "a" * 128
    long_location_name = "b" * 128

    assert len(unique_docker_resource_name(long_deployment_name, long_location_name)) == 63


def test_container_kwargs_stop_timeout():
    with docker_instance() as instance:
        assert not instance.user_code_launcher._get_standalone_dagster_server_handles_for_location(  # pyright: ignore[reportAttributeAccessIssue]
            deployment_name="foo",
            location_name="bar",
        )

        result = instance.user_code_launcher._start_new_server_spinup(  # type: ignore
            deployment_name="foo",
            location_name="bar",
            desired_entry=UserCodeLauncherEntry(
                code_location_deploy_data=CodeLocationDeployData(
                    image="dagster/dagster-cloud-examples:1.9.10",
                    package_name="dagster-cloud-examples",
                    container_context={"docker": {"container_kwargs": {"stop_timeout": 23}}},
                ),
                update_timestamp=time.time(),
            ),
        )
        assert result.server_handle.container.labels[STOP_TIMEOUT_LABEL] == "23"

        assert instance.user_code_launcher._get_standalone_dagster_server_handles_for_location(  # pyright: ignore[reportAttributeAccessIssue]
            deployment_name="foo",
            location_name="bar",
        )

        instance.user_code_launcher._remove_server_handle(result.server_handle)  # type: ignore



from dagster._core.libraries import DagsterLibraryRegistry

from dagster_cloud.instance import DagsterCloudAgentInstance as DagsterCloudAgentInstance
from dagster_cloud.storage.compute_logs import CloudComputeLogManager as CloudComputeLogManager
from dagster_cloud.version import __version__

DagsterLibraryRegistry.register("dagster-cloud", __version__)



from dagster import AssetKey, SourceAsset, asset
from dagster._core.definitions.metadata import (
    CodeReferencesMetadataSet,
    CodeReferencesMetadataValue,
    LocalFileCodeReference,
)

# importing this makes it show up twice when we collect everything
from dagster_cloud_tests.metadata_tests.asset_package.asset_subpackage.another_module_with_assets import (
    miles_davis,
)

assert miles_davis

elvis_presley = SourceAsset(key=AssetKey("elvis_presley"))


@asset(
    metadata={
        **CodeReferencesMetadataSet(
            code_references=CodeReferencesMetadataValue(
                code_references=[LocalFileCodeReference(file_path=__file__, line_number=1)]
            )
        ),
    }
)
def chuck_berry(elvis_presley, miles_davis):
    pass



import json
import os
from unittest import mock

import pytest
import yaml
from dagster_cloud.agent.cli import app
from dagster_cloud.workspace.docker import DockerUserCodeLauncher
from dagster_cloud.workspace.user_code_launcher import ProcessUserCodeLauncher
from typer.testing import CliRunner


@pytest.fixture
def token():
    return "agent:hooli:1b2c814969d746b09d9k4d7e6g065cn4"


@pytest.fixture
def deployment():
    return "staging"


@pytest.fixture
def agent_label():
    return "my-agent"


@pytest.fixture
def dagster_config(token, deployment, agent_label):
    return {
        "instance_class": {
            "module": "dagster_cloud.instance",
            "class": "DagsterCloudAgentInstance",
        },
        "dagster_cloud_api": {
            "agent_token": token,
            "deployment": deployment,
            "agent_label": agent_label,
        },
        "user_code_launcher": {
            "module": "dagster_cloud.workspace.docker",
            "class": "DockerUserCodeLauncher",
        },
    }


@pytest.fixture
def dagster_home(tmpdir, dagster_config):
    dagster_yaml = tmpdir / "dagster.yaml"
    dagster_yaml.write_text(yaml.dump(dagster_config), "utf-8")
    return str(tmpdir)


@pytest.fixture
def instance_spy():
    class Spy:
        def __init__(self):
            self._spy = None

        def set(self, obj):
            self._spy = obj

        def __getattr__(self, name):
            return getattr(self._spy, name)

    return Spy()


@pytest.fixture
def runner(instance_spy, tmpdir, monkeypatch):
    # Isolate tests from our local env
    for env in os.environ.keys():
        if env.startswith("DAGSTER_"):
            monkeypatch.setattr("os.environ", {})

    def mock_run_loop(_agent, *args, **kwargs):
        instance_spy.set(_agent._instance)  # noqa: SLF001

    with mock.patch("dagster_cloud.agent.cli.DagsterCloudAgent.run_loop", mock_run_loop):
        yield CliRunner()


# Happy paths
def test_dagster_home_env(runner, monkeypatch, dagster_home, instance_spy, token, deployment):
    monkeypatch.setenv("DAGSTER_HOME", dagster_home)
    result = runner.invoke(app, [])
    assert not result.exit_code
    assert instance_spy.dagster_cloud_agent_token == token


def test_dagster_home_arg(runner, dagster_home, instance_spy, token, deployment):
    result = runner.invoke(app, [dagster_home])
    assert not result.exit_code
    assert instance_spy.dagster_cloud_agent_token == token
    assert instance_spy.deployment_name == deployment


def test_options(runner, instance_spy, token, deployment):
    result = runner.invoke(app, ["--agent-token", token, "--deployment", deployment])
    assert not result.exit_code
    assert instance_spy.dagster_cloud_agent_token == token
    assert instance_spy.deployment_name == deployment


def test_config(runner, instance_spy, dagster_config, token, deployment):
    result = runner.invoke(app, ["--config", json.dumps(dagster_config)])
    assert not result.exit_code
    assert instance_spy.dagster_cloud_agent_token == token
    assert instance_spy.deployment_name == deployment
    assert isinstance(instance_spy.user_code_launcher, DockerUserCodeLauncher)


# Failures
def test_nothing_configured_fails(runner):
    result = runner.invoke(app, [])
    assert result.exit_code
    assert "No directory provided" in result.output


def test_options_incomplete_fails(runner, token, deployment):
    # No deployment
    result = runner.invoke(app, ["--agent-token", token])
    assert result.exit_code
    assert "both an agent token and a deployment" in result.output

    # No agent token
    result = runner.invoke(app, ["--deployment", deployment])
    assert result.exit_code
    assert "both an agent token and a deployment" in result.output


def test_options_and_dagster_home_arg_fails(runner, dagster_home, token, deployment):
    result = runner.invoke(app, [dagster_home, "--agent-token", token, "--deployment", deployment])
    assert result.exit_code
    assert "Cannot supply both" in result.output


def test_config_and_dagster_home_arg_fails(runner, dagster_home, dagster_config):
    result = runner.invoke(app, [dagster_home, "--config", dagster_config])
    assert result.exit_code
    assert "Cannot supply both" in result.output


def test_dagster_cloud_api_config_invalid(runner):
    result = runner.invoke(app, [dagster_home, "--config", json.dumps({"foo": "bar"})])
    assert result.exit_code


# Order of precedence
def test_prefer_options_over_dagster_home_env(
    runner, monkeypatch, instance_spy, dagster_config, dagster_home, token, deployment
):
    monkeypatch.setenv("DAGSTER_HOME", dagster_home)

    overload_deployment = "overload"
    assert overload_deployment != deployment

    result = runner.invoke(app, ["--agent-token", token, "--deployment", overload_deployment])
    assert not result.exit_code
    assert instance_spy.dagster_cloud_agent_token == token
    assert instance_spy.deployment_name == overload_deployment

    # Other config set in dagster.yaml are ignored
    assert dagster_config["dagster_cloud_api"]["agent_label"]
    assert not instance_spy.dagster_cloud_api_agent_label


def test_prefer_dagster_home_arg_over_env(
    runner, monkeypatch, instance_spy, dagster_home, token, deployment
):
    monkeypatch.setenv("DAGSTER_HOME", "/fake/directory")
    result = runner.invoke(app, [dagster_home])
    assert not result.exit_code
    assert instance_spy.dagster_cloud_agent_token == token
    assert instance_spy.deployment_name == deployment


def test_prefer_config_over_dagster_home_env(runner, monkeypatch, instance_spy, dagster_config):
    monkeypatch.setenv("DAGSTER_HOME", "/fake/directory")
    result = runner.invoke(app, ["--config", json.dumps(dagster_config)])
    assert not result.exit_code
    assert (
        instance_spy.dagster_cloud_agent_token == dagster_config["dagster_cloud_api"]["agent_token"]
    )
    assert instance_spy.deployment_name == dagster_config["dagster_cloud_api"]["deployment"]


def test_merge_options_into_config(runner, monkeypatch, instance_spy, dagster_config):
    overload_deployment = "overload"
    assert overload_deployment != dagster_config["dagster_cloud_api"]["deployment"]

    result = runner.invoke(
        app,
        [
            "--config",
            json.dumps(dagster_config),
            "--deployment",
            overload_deployment,
            "--user-code-launcher",
            "dagster_cloud.workspace.user_code_launcher.ProcessUserCodeLauncher",
        ],
        catch_exceptions=False,
    )
    assert not result.exit_code
    assert (
        instance_spy.dagster_cloud_agent_token == dagster_config["dagster_cloud_api"]["agent_token"]
    )
    assert instance_spy.deployment_name == overload_deployment
    assert isinstance(instance_spy.user_code_launcher, ProcessUserCodeLauncher)


# Other options
def test_agent_label(runner, instance_spy, token, deployment):
    label = "test"
    result = runner.invoke(
        app, ["--agent-token", token, "--deployment", deployment, "--agent-label", label]
    )
    assert not result.exit_code
    assert instance_spy.dagster_cloud_api_agent_label == label


def test_launcher_config(runner, instance_spy, token, deployment):
    env_vars = ["FOO", "BAR"]
    config = json.dumps({"env_vars": env_vars})
    result = runner.invoke(
        app,
        [
            "--agent-token",
            token,
            "--deployment",
            deployment,
            "--user-code-launcher",
            "dagster_cloud.workspace.docker.DockerUserCodeLauncher",
            "--user-code-launcher-config",
            config,
        ],
    )
    assert not result.exit_code
    assert isinstance(instance_spy.user_code_launcher, DockerUserCodeLauncher)
    assert instance_spy.user_code_launcher.env_vars == env_vars


@pytest.mark.skip(reason="TODO")
def test_logging_config(runner):
    pass



import os
import sys
import tempfile
from pathlib import Path
from typing import Any
from unittest import mock

import pytest
from dagster_cloud.auth.constants import get_hardcoded_test_agent_token
from dagster_cloud_cli.commands.serverless import build_python_executable_command
from dagster_cloud_cli.core.pex_builder import util
from dagster_cloud_cli.core.pex_builder.deps import BuildMethod

from dagster_cloud_tests import gen_agent_instance


def pytest_addoption(parser: pytest.Parser):
    parser.addoption(
        "--agent-batching-enabled",
        action="store_true",
        default=False,
        help="Enable agent api response batching",
    )


@pytest.fixture(scope="session", autouse=True)
def enable_agent_batching(pytestconfig):
    """Enable agent batching, iff pytest is invoked with --agent-batching-enabled."""
    batching_enabled = pytestconfig.getoption("agent_batching_enabled")
    os.environ["DAGSTER_CLOUD_AGENT_UPLOAD_API_RESPONSE_BATCHING_ENABLED"] = (
        "true" if batching_enabled else "false"
    )


@pytest.fixture(scope="module")
def agent_token():
    return get_hardcoded_test_agent_token("acme")


@pytest.fixture
def dagster_cloud_url():
    return "https://agent-api-test.dagster.cloud"


@pytest.fixture
def agent_instance(dagster_cloud_url):
    with gen_agent_instance(dagster_cloud_url, "token") as instance:
        yield instance


@pytest.fixture(scope="session")
def sample_repos_dir():
    # Returns a Path object the ./sample-repos folder
    # Usage:
    # (sample_repos_dir / 'repo-1') -> Path('/full/path/to/tmp/repo-1')
    return Path(__file__).parent / "sample-repos"


class CachingPexBuilder:
    def __init__(self, source_repos_dir, build_dir):
        self.source_repos_dir = Path(source_repos_dir)
        self.build_dir = Path(build_dir)

    def __getitem__(self, repo_name: str):
        repo_dir = self.source_repos_dir / repo_name
        if not repo_dir.exists():
            raise ValueError(f"Could not find repo {repo_name} under {self.source_repos_dir}")

        repo_build_dir = self.build_dir / repo_name

        if not repo_build_dir.exists():
            self.build_pex_files(repo_dir, repo_build_dir)
        return self.files_from_repo_build_dir(repo_build_dir)

    def files_from_repo_build_dir(self, repo_build_dir: Path):
        files: dict[str, Any] = {"all": [], "deps": [], "source": None}
        for path in repo_build_dir.iterdir():
            if path.name == ".pex" or not path.name.endswith(".pex"):
                continue
            files["all"].append(path)
            if path.name.startswith("source-"):
                files["source"] = path
            else:
                files["deps"].append(path)

        files["pex_tag"] = "files=" + ":".join(sorted(filepath.name for filepath in files["all"]))
        return files

    def build_pex_files(self, repo_source_dir, repo_build_dir):
        with mock.patch("dagster_cloud_cli.commands.metrics.gql"):
            build_python_executable_command(
                code_directory=repo_source_dir,
                output_directory=repo_build_dir,
                python_version="3.12",
                api_token="fake",
                url="fake",
                build_method=BuildMethod.LOCAL,
            )


@pytest.fixture(scope="session")
def sample_repos_pex_files(sample_repos_dir, patched_pex_flags):
    with tempfile.TemporaryDirectory() as build_dir:
        yield CachingPexBuilder(sample_repos_dir, build_dir)


@pytest.fixture(scope="session")
def patched_pex_flags(session_mocker):
    # to enable local testing for pex builds on mac:
    # - if arch is linux we assume it is CICD and leave the pex flags as-is
    # - otherwise, assume local test and set the pex platform to 'current'

    orig_get_pex_flags = util.get_pex_flags

    def get_pex_flags_patched(*args, **kwargs):
        flags = orig_get_pex_flags(*args, **kwargs)
        new_flags = []
        for flag in flags:
            if flag.startswith("--platform"):
                new_flags.append("--platform=current")
            else:
                new_flags.append(flag)

        return new_flags

    if sys.platform != "linux":
        session_mocker.patch(
            "dagster_cloud_cli.core.pex_builder.util.get_pex_flags", get_pex_flags_patched
        )

    yield


@pytest.fixture
def venvs_root(monkeypatch):
    with tempfile.TemporaryDirectory() as venv_dir:
        monkeypatch.setenv("VENVS_ROOT", venv_dir)
        yield Path(venv_dir)



import os

from dagster._core.test_utils import environ
from dagster_cloud.secrets import DagsterCloudSecretsLoader
from dagster_cloud_backend.storage.host_cloud.cloud_storage.secrets import (
    DagsterCloudSecretMetadata,
    FullDeploymentSecretScope,
    SecretValue,
)
from dagster_cloud_test_infra.secrets.fixtures import mock_secret_store  # noqa: F401 (fixture)


def test_instance_secrets_loader(
    agent_instance_local_ursula,
    host_instance,
    mock_secret_store,  # noqa: F811 (fixture)
):
    agent_instance = agent_instance_local_ursula
    assert isinstance(
        agent_instance._secrets_loader,  # noqa: SLF001
        DagsterCloudSecretsLoader,
    )

    host_instance.cloud_storage.create_secret(
        DagsterCloudSecretMetadata(
            secret_name="FOO_KEY",
            deployment_scopes=[FullDeploymentSecretScope()],
            location_names=["foo_location"],
        ),
        SecretValue("bar_with_location"),
        updated_by=None,
    )

    host_instance.cloud_storage.create_secret(
        DagsterCloudSecretMetadata(
            secret_name="FOO_KEY",
            deployment_scopes=[FullDeploymentSecretScope()],
        ),
        SecretValue("bar"),
        updated_by=None,
    )

    assert "FOO_KEY" not in os.environ

    with environ(
        {"FOO_KEY": None}  # pyright: ignore[reportArgumentType]
    ):  # clean up after messing with os.environ
        agent_instance.inject_env_vars(location_name="foo_location")

        assert os.environ.get("FOO_KEY") == "bar_with_location"

    with environ(
        {"FOO_KEY": None}  # pyright: ignore[reportArgumentType]
    ):  # clean up after messing with os.environ
        agent_instance.inject_env_vars(location_name=None)

        assert os.environ.get("FOO_KEY") == "bar"



import os
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from unittest import mock

import pytest
from dagster_dbt import DbtCliResource
from google.api_core.exceptions import NotFound
from google.cloud.bigquery import DatasetReference

JAFFLE_SHOP_ASSET_KEYS = {
    "customers",
    "orders",
    "raw_customers",
    "raw_orders",
    "raw_payments",
    "stg_customers",
    "stg_orders",
    "stg_payments",
}


@dataclass
class BigqueryCostRow:
    job_id: str
    unique_id: str
    bytes_billed: int
    slots_ms: int


@pytest.fixture(scope="session")
def snowflake_jaffle_dir():
    yield Path(__file__).parent / "snowflake_jaffle"


@pytest.fixture(scope="session")
def snowflake_manifest_path(snowflake_jaffle_dir):
    dbt = DbtCliResource(project_dir=os.fspath(snowflake_jaffle_dir))
    dbt.cli(["deps"]).wait()
    dbt_parse_invocation = dbt.cli(["parse"]).wait()
    snowflake_manifest_path = dbt_parse_invocation.target_path.joinpath("manifest.json")
    yield snowflake_manifest_path


@pytest.fixture(scope="session")
def bigquery_jaffle_dir():
    yield Path(__file__).parent / "bigquery_jaffle"


@pytest.fixture(scope="session")
def bigquery_manifest_path(bigquery_jaffle_dir):
    dbt = DbtCliResource(project_dir=os.fspath(bigquery_jaffle_dir))
    dbt.cli(["deps"]).wait()
    dbt_parse_invocation = dbt.cli(["parse"]).wait()
    manifest_path = dbt_parse_invocation.target_path.joinpath("manifest.json")
    yield manifest_path


@contextmanager
def bigquery_client(
    client_location: str | None = None,
    client_project: str | None = None,
    dataset_location: str | None = None,
    dataset_project: str | None = None,
):
    with mock.patch(
        "dbt.adapters.contracts.connection.Connection.handle",
        new_callable=mock.PropertyMock,
    ) as mock_get_client:
        mock_dataset = mock.MagicMock()
        mock_dataset.location = dataset_location
        mock_dataset.project = dataset_project
        mock_client = mock.MagicMock()
        mock_client.get_dataset.return_value = mock_dataset
        mock_client.query.return_value = _fake_cost_rows()
        mock_client.location = client_location
        mock_client.project = client_project
        mock_get_client.return_value = mock_client
        yield mock_client


def _fake_cost_rows():
    rows = []
    for idx, asset_key_string in enumerate(JAFFLE_SHOP_ASSET_KEYS):
        unique_id = (
            f"seed.jaffle_shop.{asset_key_string}"
            if asset_key_string.startswith("raw_")
            else f"model.jaffle_shop.{asset_key_string}"
        )
        # make sure we have a mix of non-null and null values for bytes_billed and slot_ms
        if idx % 3 == 0:
            bytes_billed = 1000
            slot_ms = 1
        elif idx % 3 == 1:
            bytes_billed = None
            slot_ms = 1
        else:
            bytes_billed = 1000
            slot_ms = None
        rows.append(BigqueryCostRow("fake_job_id", unique_id, bytes_billed, slot_ms))  # pyright: ignore[reportArgumentType]
    return rows


@pytest.fixture
def default_bigquery_client():
    with bigquery_client("US", "fake_project") as client:
        yield client


@pytest.fixture(scope="session")
def bigquery_with_check_failure_jaffle_dir():
    yield Path(__file__).parent / "bigquery_jaffle_with_check_failure"


@pytest.fixture(scope="session")
def bigquery_with_check_failure_manifest_path(bigquery_with_check_failure_jaffle_dir):
    dbt = DbtCliResource(project_dir=os.fspath(bigquery_with_check_failure_jaffle_dir))
    dbt.cli(["deps"]).wait()
    dbt_parse_invocation = dbt.cli(["parse"]).wait()
    manifest_path = dbt_parse_invocation.target_path.joinpath("manifest.json")
    yield manifest_path


def get_dataset_side_effect(dataset_ref):
    if isinstance(dataset_ref, DatasetReference) and dataset_ref.project == "fake_default_project":
        mock_dataset = mock.MagicMock()
        mock_dataset.location = "US"
        mock_dataset.project = "fake_default_project"
        return mock_dataset
    else:
        raise NotFound("Dataset not found")


def create_dataset_ref_side_effect(schema, project):
    return DatasetReference(project, schema)


@pytest.fixture
def bigquery_client_with_execution_project():
    with mock.patch(
        "dbt.adapters.contracts.connection.Connection.handle",
        new_callable=mock.PropertyMock,
    ) as mock_get_client:
        mock_client = mock.MagicMock()
        mock_client.get_dataset = mock.MagicMock(side_effect=get_dataset_side_effect)
        mock_client.query.return_value = _fake_cost_rows()
        mock_client.location = None
        mock_client.project = "fake_exec_project"
        mock_client.dataset = mock.MagicMock(side_effect=create_dataset_ref_side_effect)
        mock_get_client.return_value = mock_client
        yield mock_client


@pytest.fixture(scope="session")
def bigquery_with_execution_project_jaffle_dir():
    yield Path(__file__).parent / "bigquery_jaffle_with_execution_project"


@pytest.fixture(scope="session")
def bigquery_with_execution_project_manifest_path(bigquery_with_execution_project_jaffle_dir):
    dbt = DbtCliResource(project_dir=os.fspath(bigquery_with_execution_project_jaffle_dir))
    dbt.cli(["deps"]).wait()
    dbt_parse_invocation = dbt.cli(["parse"]).wait()
    manifest_path = dbt_parse_invocation.target_path.joinpath("manifest.json")
    yield manifest_path



import pytest


def pytest_addoption(parser: pytest.Parser):
    parser.addoption("--force-shard1", action="store_true", default=False, help="Enable sharding")


pytest_plugins = [
    "dagster_cloud_test_infra.instance_fixtures",
]
